package ims.data;

import java.util.ArrayList;
import java.util.List;

public class CoNLL06Sentence extends AbstractSentence {

	protected final String[] lemma;
	protected final String[] cpos;
	protected final String[] pos;
	protected final String[] feat;
	protected final String[][] splitFeat;
	protected final int[] head;
	protected final String[] lbl;
	protected final int[] projHead;
	protected final String[] projLbl;
	
	
	public CoNLL06Sentence(int sentId,String[] id, String[] form, String[] lemma, String[] cpos, String[] pos,String[] feat,int[] head,String[] lbl,int[] projHead,String[] projLbl) {
		super(sentId, id, form);
		this.lemma=lemma;
		this.cpos=cpos;
		this.pos=pos;
		this.feat=feat;
		this.head=head;
		this.lbl=lbl;
		this.projHead=projHead;
		this.projLbl=projLbl;
		splitFeat=new String[id.length][];
		for(int i=0;i<id.length;++i)
			splitFeat[i]=this.feat[i].split(BAR_REGEXP);
	}


	@Override
	public String[] getTrainLemma() {
		return lemma;
	}
	
	@Override
	public String[] getTrainPOS() {
		return pos;
	}

	@Override
	public String[][] getTrainMorph() {
		return splitFeat;
	}

	@Override
	public int[] getTrainHeads() {
		return head;
	}

	@Override
	public String[] getTrainLabels() {
		return lbl;
	}

	@Override
	public void setPredHeads(int[] heads) {
		if(this.head.length!=heads.length)
			throw new Error("wrong length");
		for(int i=0;i<heads.length;++i)
			this.head[i]=heads[i];
	}

	@Override
	public void setPredLabels(String[] labels) {
		if(this.lbl.length!=labels.length)
			throw new Error("wrong length");
		for(int i=0;i<lbl.length;++i)
			this.lbl[i]=labels[i];
	}

	@Override
	public int[] getGoldHeads() {
		return head;
	}

	@Override
	public String[] getGoldLabels() {
		return lbl;
	}

	@Override
	public int[] getPredHeads() {
		return head;
	}


	@Override
	public String[] getPredLabels() {
		return lbl;
	}
	
	
	@Override
	public List<String> getTabbedLineList() {
		List<String> lines=new ArrayList<>();
		for(int i=1;i<form.length;++i){
			StringBuilder sb=new StringBuilder();
			sb.append(id[i]).append('\t');
			sb.append(form[i]).append('\t');
			sb.append(lemma[i]).append('\t');
			sb.append(cpos[i]).append('\t');
			sb.append(pos[i]).append('\t');
			sb.append(feat[i]).append('\t');
			if(head[i]==-1)	sb.append("_").append('\t');
			else sb.append(head[i]).append('\t');
			sb.append(lbl[i]).append('\t');
			if(projHead[i]==-1)	sb.append("_").append('\t');
			else sb.append(projHead[i]).append('\t');
			sb.append(projLbl[i]);			
			lines.add(sb.toString());
		}
		return lines;
	}

	public static CoNLL06Sentence lines2Sentence(List<String> lines,Integer sid) {
		int size=lines.size()+1; //one for root
		String[] id=new String[size];
		String[] form=new String[size];
		String[] lemma=new String[size];
		String[] cpos=new String[size];
		String[] pos=new String[size];
		String[] feat=new String[size];
		int[] head=new int[size];
		String[] lbl=new String[size];
		int[] projHead=new int[size];
		String[] projLbl=new String[size];

		form[0]=ISentence.ROOT;
		lemma[0]=ISentence.ROOT;
		cpos[0]=ISentence.ROOT;
		pos[0]=ISentence.ROOT;
		feat[0]=ISentence.ROOT;
		head[0]=ISentence.ROOT_HEAD;
		lbl[0]=ISentence.ROOT;
		projHead[0]=ISentence.ROOT_HEAD;
		projLbl[0]=ISentence.ROOT;
		
		int i=1;
		for(String l:lines){
//			System.out.println(l);
			String[] cols=l.split("\\s+");
			id[i]=cols[0];
			form[i]=cols[1];
			lemma[i]=cols[2];
			cpos[i]=cols[3];
			pos[i]=cols[4];
			feat[i]=cols[5];
			head[i]=cols[6].equals("_")?-1:Integer.parseInt(cols[6]);
			lbl[i]=cols[7];
			projHead[i]=cols[8].equals("_")?-1:Integer.parseInt(cols[8]);
			projLbl[i]=cols[9];
			++i;
		}
		
		CoNLL06Sentence s=new CoNLL06Sentence(sid,id,form,lemma,cpos,pos,feat,head,lbl,projHead,projLbl);
		return s;
	}


	@Override
	public String[] getStag() {
		return null;
	}


	@Override
	public int[] getStackHead() {
		return null;
	}


	@Override
	public String[] getStackLbl() {
		return null;
	}


	@Override
	public String[] getGoldPOS() {
		return pos;
	}


	@Override
	public String[] getFullTrainMorph() {
		return feat;
	}

	@Override
	public void reverseSentence() {
		super.reverseSentence();
		reverseObj(lemma);
		reverseObj(cpos);
		reverseObj(pos);
		reverseObj(feat);
		reverseObj(splitFeat);
		reverseObj(lbl);
		reverseObj(projLbl);
		reverseHeads(head);
		reverseHeads(projHead);
	}


	@Override
	public void insertTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.insertTokenOffsetInID(tokOffset);
		super.offsetHeadIntIndex(head, tokOffset);
		super.offsetHeadIntIndex(projHead, tokOffset);
	}


	@Override
	public void renormalizeTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.renormalizeTokenOffsetInID(tokOffset);
		super.renormalizeHeadIntIndex(head, tokOffset);
		super.renormalizeHeadIntIndex(projHead, tokOffset);
	}


	@Override
	public String getDocID() {
		throw new Error("doesn't work");
	}


	@Override
	public ISentence getSubSentence(int b, int e, int tokOffset) {
		throw new Error("too lazy to implement");
	}

}
