package ims.data;

import java.util.ArrayList;
import java.util.List;

public class CoNLL09Sentence extends AbstractSentence{

	public final String[] lemma;
	public final String[] plemma;
	public final String[] pos;
	public final String[] ppos;
	public final String[] feat;
	public final String[][] splitFeat;
	public final String[] pfeat;
	public final String[][] splitPFeat;
	public final int[] head;
	public final int[] phead;
	public final String[] lbl;
	public final String[] plbl;
	public final String[][] rest;
	
	
	protected CoNLL09Sentence(int sentId, String[] ids, String[] forms,String[] lemma, String[] plemma,String[] pos, String[] ppos, String[] feat, String[] pfeat, int[] head, String[] lbl,int[] phead, String[] plbl,String[][] rest) {
		super(sentId, ids, forms);
		this.lemma=lemma;
		this.plemma=plemma;
		this.pos=pos;
		this.ppos=ppos;
		this.feat=feat;
		this.pfeat=pfeat;
		this.head=head;
		this.lbl=lbl;
		this.phead=phead;
		this.plbl=plbl;
		this.rest=rest;
		
		this.splitFeat=new String[ids.length][];
		this.splitPFeat=new String[ids.length][];
		for(int i=0;i<ids.length;++i){
			this.splitFeat[i]=this.feat[i].split(BAR_REGEXP);
			this.splitPFeat[i]=this.pfeat[i].split(BAR_REGEXP);
		}
	}

	@Override
	public String[] getTrainLemma() {
		return plemma;
	}
	
	@Override
	public String[] getTrainPOS() {
		return ppos;
	}

	@Override
	public String[][] getTrainMorph() {
		return splitPFeat;
	}

	@Override
	public int[] getTrainHeads() {
		return head;
	}

	@Override
	public String[] getTrainLabels() {
		return lbl;
	}

	@Override
	public void setPredHeads(int[] heads) {
		if(heads.length!=this.phead.length)
			throw new Error("wrong length");
		for(int i=0;i<heads.length;++i)
			this.phead[i]=heads[i];
	}

	@Override
	public void setPredLabels(String[] labels) {
		if(labels.length!=this.plbl.length)
			throw new Error("wrong length");
		for(int i=0;i<labels.length;++i)
			this.plbl[i]=labels[i];
	}

	@Override
	public int[] getGoldHeads() {
		return head;
	}

	@Override
	public String[] getGoldLabels() {
		return lbl;
	}

	@Override
	public int[] getPredHeads() {
		return phead;
	}

	@Override
	public String[] getPredLabels() {
		return plbl;
	}
	
	
	@Override
	public List<String> getTabbedLineList() {
		List<String> lines=new ArrayList<>();
		for(int i=1;i<form.length;++i){
			StringBuilder sb=new StringBuilder();
			sb.append(id[i]).append('\t');
			sb.append(form[i]).append('\t');
			sb.append(lemma[i]).append('\t');
			sb.append(plemma[i]).append('\t');
			sb.append(pos[i]).append('\t');
			sb.append(ppos[i]).append('\t');
			sb.append(feat[i]).append('\t');
			sb.append(pfeat[i]).append('\t');
			if(head[i]<0) 
				sb.append("_").append('\t');
			else sb.append(head[i]).append('\t');
			if(phead[i]<0) sb.append("_").append('\t');
			else sb.append(phead[i]).append('\t');
			sb.append(lbl[i]).append('\t');
			sb.append(plbl[i]);
			for(String s:rest[i])
				sb.append('\t').append(s);			
			lines.add(sb.toString());
		}
		return lines;
	}
	
	public static CoNLL09Sentence lines2Sentence(List<String> lines,Integer sid) {
		int size=lines.size()+1; //one for root
		String[] id=new String[size];
		String[] form=new String[size];
		String[] lemma=new String[size];
		String[] plemma=new String[size];
		String[] pos=new String[size];
		String[] ppos=new String[size];
		String[] feat=new String[size];
		String[] pfeat=new String[size];
		int[] head=new int[size];
		String[] lbl=new String[size];
		int[] pHead=new int[size];
		String[] pLbl=new String[size];

		String[][] rest=new String[size][];
		
		form[0]=ISentence.ROOT;
		lemma[0]=ISentence.ROOT;
		plemma[0]=ISentence.ROOT;
		pos[0]=ISentence.ROOT;
		ppos[0]=ISentence.ROOT;
		feat[0]=ISentence.ROOT;
		pfeat[0]=ISentence.ROOT;
		head[0]=ISentence.ROOT_HEAD;
		lbl[0]=ISentence.ROOT;
		pHead[0]=ISentence.ROOT_HEAD;
		pLbl[0]=ISentence.ROOT;
		
		int i=1;
		for(String l:lines){
//			System.out.println(l);
			String[] cols=l.split("\\s+");
			id[i]=cols[0];
			form[i]=cols[1];
			lemma[i]=cols[2];
			plemma[i]=cols[3];
			pos[i]=cols[4];
			ppos[i]=cols[5];
			feat[i]=cols[6];
			pfeat[i]=cols[7];
			head[i]=cols[8].equals("_")?-1:Integer.parseInt(cols[8]);
			pHead[i]=cols[9].equals("_")?-1:Integer.parseInt(cols[9]);
			lbl[i]=cols[10];
			pLbl[i]=cols[11];
			rest[i]=new String[cols.length-12];
			for(int c=12,q=0;c<cols.length;++c,++q)
				rest[i][q]=cols[c];
			++i;
		}
		
		CoNLL09Sentence s=new CoNLL09Sentence(sid,id,form,lemma,plemma,pos,ppos,feat,pfeat,head,lbl,pHead,pLbl,rest);
		return s;
	}

	@Override
	public String[] getStag() {
		return null;
	}

	@Override
	public int[] getStackHead() {
		return null;
	}

	@Override
	public String[] getStackLbl() {
		return null;
	}

	@Override
	public String[] getGoldPOS() {
		return pos;
	}

	@Override
	public String[] getFullTrainMorph() {
		return pfeat;
	}
	
	@Override
	public void reverseSentence() {
		super.reverseSentence();
		reverseObj(lemma);
		reverseObj(plemma);
		reverseObj(pos);
		reverseObj(ppos);
		reverseObj(feat);
		reverseObj(splitFeat);
		reverseObj(pfeat);
		reverseObj(splitPFeat);
		reverseHeads(head);
		reverseHeads(phead);
		reverseObj(lbl);
		reverseObj(plbl);
		reverseObj(rest);
	}

	@Override
	public void insertTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.insertTokenOffsetInID(tokOffset);
		super.offsetHeadIntIndex(head, tokOffset);
		super.offsetHeadIntIndex(phead, tokOffset);		
	}

	@Override
	public void renormalizeTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.renormalizeTokenOffsetInID(tokOffset);
		super.renormalizeHeadIntIndex(head, tokOffset);
		super.renormalizeHeadIntIndex(phead, tokOffset);
	}

	@Override
	public String getDocID() {
		return rest[1][rest[1].length-1];
	}

	@Override
	public CoNLL09Sentence getSubSentence(int b, int e, int tokOffset) {
		int size=e-b+1; //one for root
		String[] id=new String[size];
		String[] form=new String[size];
		String[] lemma=new String[size];
		String[] plemma=new String[size];
		String[] pos=new String[size];
		String[] ppos=new String[size];
		String[] feat=new String[size];
		String[] pfeat=new String[size];
		int[] head=new int[size];
		String[] lbl=new String[size];
		int[] pHead=new int[size];
		String[] pLbl=new String[size];

		String[][] rest=new String[size][];
		
		form[0]=ISentence.ROOT;
		lemma[0]=ISentence.ROOT;
		plemma[0]=ISentence.ROOT;
		pos[0]=ISentence.ROOT;
		ppos[0]=ISentence.ROOT;
		feat[0]=ISentence.ROOT;
		pfeat[0]=ISentence.ROOT;
		head[0]=ISentence.ROOT_HEAD;
		lbl[0]=ISentence.ROOT;
		pHead[0]=ISentence.ROOT_HEAD;
		pLbl[0]=ISentence.ROOT;

		System.arraycopy(this.id, b, id, 1, e-b);
		System.arraycopy(this.form, b, form, 1, e-b);
		System.arraycopy(this.lemma, b, lemma, 1, e-b);
		System.arraycopy(this.plemma, b, plemma, 1, e-b);
		System.arraycopy(this.pos, b, pos, 1, e-b);
		System.arraycopy(this.ppos, b, ppos, 1, e-b);
		System.arraycopy(this.feat, b, feat, 1, e-b);
		System.arraycopy(this.pfeat, b, pfeat, 1, e-b);
		System.arraycopy(this.head, b, head, 1, e-b);
		System.arraycopy(this.phead, b, pHead, 1, e-b);
		System.arraycopy(this.lbl, b, lbl, 1, e-b);
		System.arraycopy(this.plbl, b, pLbl, 1, e-b);
		System.arraycopy(this.rest, b, rest, 1, e-b);
		

		CoNLL09Sentence r=new CoNLL09Sentence(-1, id, form, lemma, plemma, pos, ppos, feat, pfeat, head, lbl, pHead, pLbl, rest);
		r.renormalizeTokenOffsetsInHeadAndIDCols(tokOffset);
		return r;
	}

}
