package ims.io;


import java.io.IOException;
import java.util.Iterator;

import ims.data.ISentence;
import ims.util.AbstractSingleItemBufferedIterator;
import ims.util.NonProjInfo;

public class NonProjDiscardingReader<T extends ISentence> implements SentenceReader<T>{

	private final Iterable<T> reader;
	
	public NonProjDiscardingReader(SentenceReader<T> reader){
		this.reader=reader;
	}
	
	public static <T extends ISentence> NonProjDiscardingReader<T> createReader(SentenceReader<T> reader){
		return new NonProjDiscardingReader<T>(reader);
	}
	
	@Override
	public Iterator<T> iterator() {
		return new SentenceIterator(reader.iterator());
	}
	
	class SentenceIterator extends AbstractSingleItemBufferedIterator<T> {

		private final Iterator<T> it;
		
		SentenceIterator(Iterator<T> it){
			this.it=it;
			try {
				super.initSetNext(init());
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		@Override
		protected T readNext() throws IOException {
			if(!it.hasNext())
				return null;
			do {
				T s=it.next();
				int[] head=s.getGoldHeads();
				if(NonProjInfo.isProjectiveSentence(head))
					return s; 
			} while(it.hasNext());
			return null;
		}

		@Override
		protected T init() throws IOException {
			return readNext();
		}
		
	}

}