package ims.io;

import ims.data.ISentence;
import ims.trans.Options;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

public class ReaderWriterFactory {

	public static SentenceReader<?> getTrainReader(Options options) {
		return getReader(options.train,options.format);
	}
	
	public static SentenceReader<?> getReader(File input, Format format) {
		return getReader(input,format.sentenceClass);
	}
	
	public static <T extends ISentence> SentenceReader<T> getReader(File input,Class<T> sentenceClass){
		return getReader(input,sentenceClass,false);
		
	}
	
	public static <T extends ISentence> SentenceReader<T> getReader(File input,Class<T> sentenceClass,boolean reversed){
		if(reversed)
			return new ReversingReader<T>(new TabbedReader<T>(input,sentenceClass));
		else
			return new TabbedReader<T>(input,sentenceClass);
	}
	
	public static TabbedWriter getWriter(File output) throws UnsupportedEncodingException, FileNotFoundException{
		return getWriter(output,false);
	}
	
	public static TabbedWriter getWriter(File output,boolean reversed) throws UnsupportedEncodingException, FileNotFoundException{
		if(reversed)
			return new ReversingTabbedWriter(output);
		else
			return new TabbedWriter(output);
	}
}
