package ims.oracle;

import ims.oracle.Configuration.TR;
import ims.oracle.DeterministicArcStandardSwapOldOracle.InOrderOrder;
import ims.trans.Options.TransitionSystem;
import ims.util.DisjointSetZeroToN;
import ims.util.DisjointSetZeroToNRankPathComp;
import ims.util.NonProjInfo;

import java.util.ArrayList;
import java.util.List;

public class DeterministicArcStandardSwapNewOracle extends AbstractStaticOracle{

	private InOrderOrder inOrder;
	private DisjointSetZeroToN mpcs;
	
	protected DeterministicArcStandardSwapNewOracle(TR[] myTRs) {
		super(myTRs);
	}
	
	@Override
	public void initSentence(int[] heads){ //Need this for swap stuff
		inOrder=new InOrderOrder(InOrderOrder.getSequence(heads));
		mpcs=getMPCs(heads,new Configuration(heads));
	}
	
	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels, boolean labeled) {
		boolean nonProj=!NonProjInfo.isProjectiveSentence(heads);
		InOrderOrder inOrder;
		DisjointSetZeroToN mpcs; //need to track the minimal projective components (see Nivre et al., IWPT 2009). we use a union-find data structure to track that (see getMPCs())
		if(nonProj){
			inOrder=new InOrderOrder(InOrderOrder.getSequence(heads));
			mpcs=getMPCs(heads,new Configuration(heads));
		} else {
			return DeterministicArcStandardOracle.staticGetCanonicalSequence(heads, labels, labeled);
		}
		Configuration c=new Configuration(heads);
		List<String> operations=new ArrayList<String>();
		while(!(c.bufferSize()==0 && c.stackSize()==1)){
			final String op;
			if(DeterministicArcStandardOracle.canLeftArc(heads,c)){
				op=(labeled?TR.LA+"_"+labels[c.peekStack(1)]:TR.LA.toString());
				c.doLeftArc();
			} else if(DeterministicArcStandardOracle.canRightArc(heads,c)){
				op=(labeled?TR.RAS+"_"+labels[c.peekStack(0)]:TR.RAS.toString());
				c.doRightArcS();
			} else if(nonProj && canSwap(heads,c,inOrder,mpcs)){
				op=TR.SW.toString();
				c.doSwap();
			} else {
				op=TR.SH.toString();
				c.doShift();
			}
//			System.out.println(op);
			operations.add(op);
		}
		return operations;
	}

	static boolean canSwap(int[] heads, IConfiguration c, InOrderOrder inOrder, DisjointSetZeroToN mpcs) {
		if(!c.hasMinTwoStackElements())
			return false;
		int s1=c.peekStack(1);
		int s0=c.peekStack(0);
		int b0=(c.bufferSize()==0?-1:c.peekBuffer(0));
		return canSwap(s1,s0,b0,inOrder,mpcs);
	}
	
	static boolean canSwap(int s1,int s0,int b0,InOrderOrder inOrder, DisjointSetZeroToN mpcs){
		boolean inOrderHolds=inOrder.precedes(s0, s1);
		if(b0<0){
			if(inOrderHolds)
				return true;
			else
				return false;
		} else {
			boolean sameMPC=mpcs.find(s0)==mpcs.find(b0);
			if(inOrderHolds && !sameMPC)
				return true;
			else
				return false;
		}
	}

	static DisjointSetZeroToN getMPCs(int[] heads,Configuration c) {
		while(!(c.bufferSize()==0 && c.stackSize()==1)){
			if(DeterministicArcStandardOracle.canLeftArc(heads, c)){
				c.doLeftArc();
			} else if(DeterministicArcStandardOracle.canRightArc(heads, c)){
				c.doRightArcS();
			} else {
				if(c.bufferSize()==0) //So, when the buffer is empty we should stop shifting.
					break;
				else
					c.doShift();
			}
		}
		DisjointSetZeroToN mpcs=new DisjointSetZeroToNRankPathComp(heads.length);
		int[] predHeads=c.getPredHeads();
		for(int i=1;i<predHeads.length;++i)
			if(predHeads[i]>=0)
				mpcs.union(i, predHeads[i]);
		return mpcs;
	}

	@Override
	public boolean canDoNonProj() {
		return true;
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, int[] heads) {
		switch(tr){
		case LA:  return DeterministicArcStandardOracle.canLeftArc(heads, conf);
		case RAS: return DeterministicArcStandardOracle.canRightArc(heads, conf);
		case SW:  return canSwap(heads,conf,inOrder,mpcs);
		case SH:  return !DeterministicArcStandardOracle.canLeftArc(heads, conf) && !DeterministicArcStandardOracle.canRightArc(heads, conf) && !canSwap(heads,conf,inOrder,mpcs);
		default: throw new Error("tr "+tr+" not supported by this oracle");
		}
	}
	
//	@Override
//	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf, Instance inst,SymbolTable symTab) {
//		if(tr==TR.SH){
//			if(DeterministicArcStandardOracle.canLeftArc(inst.head, conf) || DeterministicArcStandardOracle.canRightArc(inst.head, conf) || canSwap(inst.head,conf,inOrder,mpcs))
//				return false;
//			else
//				return true;
//		}
//		switch(tr){
//		case LA:  return DeterministicArcStandardOracle.canLeftArc(inst.head, conf);
//		case RAS: return DeterministicArcStandardOracle.canRightArc(inst.head, conf);
//		case SW:  return canSwap(inst.head,conf,inOrder,mpcs);
//		default:  throw new Error("not supported");
//		}
//	}

	@Override
	public TransitionSystem getTS() {
		return TransitionSystem.ArcSNewSwap;
	}
}
