package ims.oracle;

import ims.data.Instance;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.StringSymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.trans.transitions.ITransition;
import ims.trans.util.FileTRSeq;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

public class DeterministicFileOracle extends AbstractStaticOracle {

	private final boolean hasSwap;
	
	protected DeterministicFileOracle(boolean withSwap) {
		super(TransitionSystem.ArcSNewSwap.trs);
		this.hasSwap=withSwap;
	}

	
	private final IOracle detOracle=OracleFactory.getStaticOracle(Options.TransitionSystem.ArcSNewSwap);
	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels,boolean labeled) {
		return detOracle.getCanonicalSequence(heads, labels, labeled);
	}

	@Override
	public boolean canDoNonProj() {
		return hasSwap;
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf,	int[] heads) {
		int i=conf.getTotalTransitions(); //correct transition will be at this position in the cachedCanonicalSeq
		if(i>=curTRSeq.length)
			return false;
//		String correct=curTRSeq[conf.getTransitionSeqSignature().length()-1];
		String correct=curTRSeq[conf.getTotalTransitions()];
		return correct.equals(tr.toString());
	}
	
	private String[] curTRSeq;
	
	@Override
	public void initSentence(Instance inst) {
		this.curTRSeq=inst.fileTrSeq;
	}
	
	@Override
	public void initSentence(int[] heads) {
		
	}

	@Override
	public TransitionSystem getTS() {
		throw new Error("not implemented");
	}
	
	public void initSeqsFromFile(File trSeqFile,List<Instance> instances,SymbolTable symTab) throws IOException{
		Iterator<List<String>> it=FileTRSeq.getSeqIt(trSeqFile);
		for(Instance inst:instances){
			if(!it.hasNext())
				throw new Error("More training instances than TR seqs in file!?");
			List<String> l=it.next();
			sanityCheckAndSetFileTRSeq(l,inst,symTab.transitions,symTab.transition2label);
		}
	}

	private void sanityCheckAndSetFileTRSeq(List<String> l, Instance inst, StringSymbolMapping transitions, ITransition[] transition2label) {
		int[] r=new int[l.size()];
		String[] p=new String[l.size()];
		MinimalConfiguration c=new MinimalConfiguration(inst.head);
		for(int i=0;i<r.length;++i){
			String op=l.get(i);
			int x=transitions.lookup(op);
			if(x==AbstractSymbolMapping.UNK_IDX)
				throw new Error("tr not in symtab!?");
			r[i]=x;
			p[i]=op.contains("_")?op.split("_", 2)[0]:op;
			ITransition tr=transition2label[x];
			c.applyTR(tr.getTR());
		}
		if(!Arrays.equals(c.predHeads,inst.head))
			throw new Error("mismatch");
		inst.fileTrSeq=p;
		inst.fileTrSeqMapped=r;
	}

	public static IOracle create(Options options) {
		try {
			Iterator<List<String>> it=FileTRSeq.getSeqIt(options.trSeqFile);
			while(it.hasNext()){
				List<String> l=it.next();
				if(l.contains(TR.SW.toString()))
					return new DeterministicFileOracle(true);
			}
		} catch (Exception e){
			throw new RuntimeException(e);
		}
		return new DeterministicFileOracle(false);
	}

}
