package ims.oracle;

import ims.oracle.Configuration.TR;
import ims.oracle.DeterministicArcStandardSwapOldOracle.InOrderOrder;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;

import java.util.List;

public class DynamicArcStandardSwapOracle extends AbstractDynamicOracle {

	/*
	 * Partial oracle wrt to {SH,SW} ambiguity. All other ambiguities are ignored.
	 * Only expensive operation is checking if SH is OK when SW also is (worst time n^2 time, but typically much less)
	 * 
	 */
	
	protected DynamicArcStandardSwapOracle(TR[] myTRs) {
		super(myTRs);
	}

	private InOrderOrder inOrder;
	
	@Override
	public void initSentence(int[] heads){ //Need this for swap stuff
		inOrder=new InOrderOrder(InOrderOrder.getSequence(heads));
	}
	
//	@Override
//	List<TR> getActions(IConfiguration c, int[] heads) {
//		if(DeterministicArcStandardOracle.canLeftArc(heads, c))
//			return Collections.singletonList(TR.LA);
//		if(DeterministicArcStandardOracle.canRightArc(heads, c))
//			return Collections.singletonList(TR.RAS);
//		boolean canSwap=DeterministicArcStandardSwapOldOracle.canSwap(heads, c, inOrder);
//		if(canSwap){ //Can we also shift?
//			boolean canShift=BruteForceShiftOKCheck.canShift(c, inOrder, heads);
//			if(canShift)
//				return Arrays.asList(TR.SH,TR.SW);
//			else
//				return Collections.singletonList(TR.SW);
//		} else {
//			return Collections.singletonList(TR.SH);
//		}
//	}

//	@Override
//	public boolean isPermissible(IConfiguration c, TR tr, int[] heads) {
//		boolean canLA=DeterministicArcStandardOracle.canLeftArc(heads, c);
//		boolean canRA=DeterministicArcStandardOracle.canRightArc(heads, c);
//		boolean canSW=DeterministicArcStandardSwapOldOracle.canSwap(heads, c, inOrder);
//		boolean canSH;
//		if(canLA || canRA)
//			canSH=false;
//		else if(canSW)
//			canSH=BruteForceShiftOKCheck.canShift(c, inOrder, heads);
//		else
//			canSH=true;
//
//		switch(tr){
//		case LA:  return canLA;
//		case RAS: return canRA;
//		case SW:  return canSW;
//		case SH:  return canSH;
//		default:  throw new Error("!");
//		}
//	}

	private final IOracle detOracle=OracleFactory.getStaticOracle(Options.TransitionSystem.ArcSNewSwap);
	@Override
	public List<String> getCanonicalSequence(int[] heads, String[] labels,boolean labeled) {
		return detOracle.getCanonicalSequence(heads, labels, labeled);
	}

	@Override
	public boolean canDoNonProj() {
		return true;
	}

	@Override
	public boolean isCorrectULBLTransition(TR tr, IConfiguration conf,int[] heads) {
		if(tr==TR.SH){
			if(DeterministicArcStandardOracle.canLeftArc(heads, conf) || DeterministicArcStandardOracle.canRightArc(heads, conf))
				return false;
			if(DeterministicArcStandardSwapOldOracle.canSwap(heads, conf, inOrder))
				return BruteForceShiftOKCheck.canShift(conf, inOrder, heads);
			else
				return conf.bufferSize()>0;
		}
		switch(tr){
		case LA:  return DeterministicArcStandardOracle.canLeftArc(heads, conf);
		case RAS: return DeterministicArcStandardOracle.canRightArc(heads, conf);
		case SW:  return DeterministicArcStandardSwapOldOracle.canSwap(heads, conf, inOrder);
		default:  throw new Error("not supported");
		}
	}

	@Override
	public TransitionSystem getTS() {
		return TransitionSystem.ArcSDynSwap;
	}

}
