package ims.trans;

import java.util.ArrayList;
import java.util.List;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.AbstractDynamicOracle;
import ims.oracle.Configuration.TR;
import ims.oracle.IOracle;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFactory.LabeledAbstractTransition;

public class DynamicOracleTrainer extends AbstractBeamTrainer{

	public DynamicOracleTrainer(Options options, SymbolTable symTab,AbstractDynamicOracle oracle, ITransitionFilter transitionFilter,List<Instance> instances, FeatureExtractor featureExtractor,Long2IntInterface l2i, ParametersFloat parameters, Decoder decoder,LossFunction lossFunction,InstanceMapping instanceMapping, int goldBeam) {
		super(options, symTab, oracle, transitionFilter, instances, featureExtractor,l2i, parameters, decoder, lossFunction, instanceMapping,goldBeam);
	}
	
	boolean isCorrectAction(ITransition lastAction, State fromState, Instance inst) {
		return isCorrectAction(lastAction,fromState,inst,oracle);
	}

	public static boolean isCorrectAction(ITransition lastAction, State fromState, Instance inst, IOracle oracle) {
		boolean ok=oracle.isCorrectULBLTransition(lastAction.getTR(), fromState, inst.head);
//		boolean ok=oracle.isCorrectULBLTransition(lastAction, fromState, inst.head);
		if(!ok)
			return false;
		return checkCorrectLabel(lastAction, fromState, inst);
	}

	private static boolean checkCorrectLabel(ITransition tr,State fromState, Instance inst) {
		if(tr.isLabeled()){
			int dep=tr.getDep(fromState);
			if(((LabeledAbstractTransition) tr).l==inst.lbl[dep])
				return true;
			else
				return false;
		} else {
			return true;
		}
	}

	
	List<ITransition> getCorrectTransitions(State lastCorrect, Instance inst) {
		return getCorrectTransitions(lastCorrect,inst,transitionFilter,symTab,oracle);
	}
	
	public static List<ITransition> getCorrectTransitions(State lastCorrect, Instance inst, ITransitionFilter transitionFilter,SymbolTable symTab,IOracle oracle) {
		List<TR> correctULBL=oracle.getActions(lastCorrect, inst.head);
		int[] t=transitionFilter.getAllowedTransitions(lastCorrect, inst);
		List<ITransition> r=new ArrayList<ITransition>();
		for(int i:t){
			ITransition tr=symTab.transition2label[i];
			if(correctULBL.contains(tr.getTR()) && checkCorrectLabel(tr,lastCorrect,inst))
				r.add(tr);
		}
		return r;
	}

}
