package ims.trans;

import java.io.IOException;
import java.util.List;

import ims.data.ISentence;
import ims.data.Instance;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.Configuration.TR;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.trans.Options.TransitionSystem;
import ims.util.Util;
import ims.util.ZipUtil;

public class GoldDynParser extends Decoder {

	public static enum GoldDecodePolicy { Greedy, Beam }
	
	public final boolean labeled;
	public final TR[] myTRs;
	public final IOracle oracle;
	public final TransitionSystem baseTs;
	private final GoldDecodePolicy goldDecodePolicy;
	private final InstanceMapping im;
	
	
	public GoldDynParser(Model model,GoldDecodePolicy goldDecodePolicy){
		super(model);
//		this.oracle=OracleFactory.getDynamicOracle(model.ts);
		this.oracle=OracleFactory.getDynamicOracle(TransitionSystem.ArcSFullDynSwap);
		this.baseTs=model.ts;
		this.labeled=model.labeled;
		this.goldDecodePolicy=goldDecodePolicy;
		this.im=model.instanceMapping;
		this.myTRs=model.ts.trs;
	}
//	
//	public void initOracle(int[] heads){
//		oracle.initSentence(heads);
//	}

	private State decode(ISentence s,int beam){
		Instance inst=im.createTestInstance(s, symTab);
		return decode(inst,beam);
	}
	
	public State decode(Instance inst,int beam){
		oracle.initSentence(inst.head);
		switch(goldDecodePolicy){
		case Greedy:         return decodeGoldGreedy(inst,oracle);
		case Beam:           return decodeGoldBeam(inst,beam,oracle);
		default:             throw new Error("not implemented: "+goldDecodePolicy);
		}
	}

	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, Options.goldModel, Model.class);
		GoldDynParser gdp=new GoldDynParser(m,Options.goldDecodePolicy);
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.train, options.format);
		for(ISentence s:reader){
			State state=gdp.decode(s,Options.goldBeam);
			List<String> trs=state.recoverTRSeqStringList();
			System.out.println(Util.joinStrings(trs, ","));
		}
	}
	
}
