package ims.trans;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.IOracle;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.ParametersFloat.IUpdStruct;
import ims.trans.AbstractBeamTrainer.Bookkeeping;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFactory.LabeledAbstractTransition;
import ims.util.DBO;
import ims.util.Pair;

import java.util.List;

public class GreedyTrainer extends AbstractTrainer {

	public GreedyTrainer(Options options, SymbolTable symTab, IOracle oracle, ITransitionFilter transitionFilter, List<Instance> instances, FeatureExtractor featureExtractor, Long2IntInterface l2i, ParametersFloat parameters, Decoder decoder, LossFunction lossFunction, InstanceMapping instanceMapping) {
		super(options, symTab, oracle, transitionFilter, instances, featureExtractor, l2i, parameters, decoder, lossFunction, instanceMapping);
	}

	@Override
	protected int doIteration(int beam, int itemCount) {
		return doGreedyIteration(itemCount);
	}
	
	private int doGreedyIteration(int itemCount) {
		long start=System.currentTimeMillis();
		IFV<FV3> goldFV=new FV3();
		IFV<FV3> predFV=new FV3();
		int erase=0;
		int correct=0;
		int total=0;
		int totalGoldSwaps=0;
		long[] sink=new long[Options.featureSinkSize];
		sink=featureExtractor.sink;
		int isCount=0;
		Bookkeeping b=new Bookkeeping();
		DBO.printNoPrefix("Sent: ");
		for(Instance is:instances){
			++isCount;
			++itemCount;
			oracle.initSentence(is);
			if(isCount%111==1)
				erase=DBO.eraseAndPrint(erase, getInfoString(isCount, total, correct,System.currentTimeMillis()-start,totalGoldSwaps));
			
			State state=State.getInitialState(is.head.length, options.labeled, is.head);
			while(!state.isTerminal()){
				++total;
//				++itemCount;
				List<Pair<ITransition,Double>> tr=decoder.getScoredTransitions(state,is);
				Pair<ITransition,Double> bestPred=extractBestPred(tr);
				final Pair<ITransition,Double> todo;
				if(isCorrectAction(state,bestPred.getLeft(),is)){
					todo=bestPred;
					++correct;
				} else {
					Pair<ITransition,Double> bestCorr=extractBestCorrect(tr,is,state);
					
					fillFeatureVectors(bestPred,bestCorr,predFV,goldFV,state,sink,is);
					
					IUpdStruct upd=parameters.getUpdStruct(bestPred.getRight(),bestCorr.getRight(),0,itemCount,goldFV.getDistVector(predFV));
					
//					PerceptronUpdStruct upd=new PerceptronUpdStruct(bestPred.getRight(),bestCorr.getRight(),0,itemCount,goldFV.getDistVector(predFV));
					parameters.update(upd);
					
					todo=bestCorr;
				}
				state=todo.getLeft().apply(state, todo.getRight());
			}
			super.bookkeepLastPredTRSeq(state, is, b);
			totalGoldSwaps+=state.countSwaps();
		}
		erase=DBO.eraseAndPrint(erase, getInfoString(isCount, total, correct,System.currentTimeMillis()-start,totalGoldSwaps));
		DBO.println();
		DBO.printlnNoPrefix(getTRSeqStats(b,isCount));
		return itemCount;
	}

	private String getInfoString(int isCount, int total, int correct, long time, int totalGoldSwaps) {
		long tpi=time/isCount;
		return String.format("%-5d  time: %d/s   TT: %-7d  CT: %-7d  GS %-7d",isCount,tpi,total,correct,totalGoldSwaps);
	}

	private void fillFeatureVectors(Pair<ITransition, Double> bestPred,	Pair<ITransition, Double> bestCorr, IFV<FV3> predFV, IFV<FV3> goldFV,State state,long[] sink,Instance is) {
		int l=featureExtractor.extractFeatures(state, sink, is);
		goldFV.clear();
		predFV.clear();
		for(int i=0;i<l;++i){
			long t=sink[i];
			t<<=featureExtractor.bA; t|=bestPred.getLeft().getId();
			predFV.add(l2i.l2i(t));
			
			long s=sink[i];
			s<<=featureExtractor.bA; s|=bestCorr.getLeft().getId();
			goldFV.add(l2i.l2i(s));
		}
	}


	private Pair<ITransition, Double> extractBestCorrect(List<Pair<ITransition, Double>> tr, Instance is, State state) {
		Pair<ITransition, Double> r=null;
		for(Pair<ITransition, Double> p:tr)
			if(isCorrectAction(state,p.getLeft(),is) && (r==null || r.getRight()<p.getRight()))
				r=p;
		return r;
	}


	private boolean isCorrectAction(State state, ITransition tr, Instance is) {
		if(!oracle.isDynamic()){
			if(tr.getId()==is.transSequence[state.totalTransitions])
				return true;
			else
				return false;
		}
		if(!oracle.isCorrectULBLTransition(tr.getTR(), state, is.head))
			return false;
		if(tr.isLabeled()){
			int dep=tr.getDep(state);
			if(((LabeledAbstractTransition) tr).l==is.lbl[dep])
				return true;
			else
				return false;
		} else {
			return true;
		}
	}


	private Pair<ITransition, Double> extractBestPred(List<Pair<ITransition, Double>> tr) {
		Pair<ITransition, Double> r=null;
		for(Pair<ITransition, Double> p:tr)
			if(r==null || r.getRight()<p.getRight())
				r=p;
		return r;
	}

}
