package ims.trans;

import ims.io.Format;
import ims.oracle.Configuration.TR;
import ims.perceptron.Regularizer.RegularizerTypes;
import ims.trans.GoldDynParser.GoldDecodePolicy;
import ims.trans.features.FeatureSet;
import ims.trans.labelfilter.LBLFilter;
import ims.util.DBO;
import ims.util.ThreadPoolSingleton;
import ims.util.Util;

import java.io.File;
import java.io.OutputStream;
import java.util.Date;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

public class Options {

	public static final int DEFAULT_BEAM_SIZE=20;

	public static enum ViolationPolicy { Early, Max, Latest, HybridMax, HybridEarly }
	
	@Option(name="-vp",aliases="-violationPolicy")
	public static ViolationPolicy violationPolicy=ViolationPolicy.Max;
	
	@Option(name="-lazy")
	public static boolean LAZY_UPDATE = false;
	
	@Option(name="-statlazy")
	public static boolean LAZY_STATIC_UPDATE = false;
	
	public final Date optionsCreate=new Date();
	
	@Option(name="-train",aliases="--train",usage="Train file")
	public File train;
	
	@Option(name="-out",aliases="--output",usage="Output file")
	public File out;

	@Option(name="-test",aliases="--test",usage="Test file")
	public File test;

	@Option(name="-gold")
	public File gold;
	
	@Option(name="-testEveryIter",aliases="-testEvery",usage="Test on test and evaluate on gold after every iteration")
	public boolean testEveryIter=false;
	
	@Option(name="-model",usage="Model file")
	public void setModel(String modelstr){
		if(modelstr.equalsIgnoreCase("null"))
			this.model=null;
		else
			this.model=new File(modelstr);
	}
	public File model=null;
	

	@Option(name="-beam",usage="beam size")
	public int beam=DEFAULT_BEAM_SIZE;
	
	@Option(name="-reg",aliases="--regularizer",usage="which PA regularizer to use")
	public RegularizerTypes regularizerType=RegularizerTypes.PA0;
	@Option(name="-C",usage="the C parameter for the regularizer (default 0.1)")
	public float C=0.1f;
	@Option(name="-hsize",aliases="--hashSize",usage="size of parameter vector for hash kernel")
//	public int hashSize=0x07ffffff; //134200007 this is a prime closeby
	public int hashSize=179424617; //this is a prime

	@Option(name="-eta",usage="set eta")
	public double eta=1.d;
	
	@Option(name="-lambda",usage="lambda")
	public double lambda=1.d;
	
	@Option(name="-iter",aliases="--iterations",usage="number of iterations for perceptron")
	public int iterations=20;
	
	@Option(name="-discardNonProj",usage="discard non-projective sentences when reading")
	public boolean discardNonProj=false;
	
	public boolean useHashKernel=true;
	@Option(name="-nohk",aliases="--noHashKernel",usage="Don't use the hash kernel -- use exact mapping")
	public void setNoHK(Boolean b){
		useHashKernel=false;
	}
	
	public boolean initBias=false;
	@Option(name="-initW",usage="init bias feature")
	public void setNoInit(Boolean b){
		initBias=true;
	}
	
	public static boolean useAveraging=true;
	@Option(name="-noavg",aliases="--noAveraging",usage="Don't use averaging")
	public void setNoAvg(Boolean b){
		useAveraging=false;
	}
	
	@Option(name="-count",usage="maximum number of training instances")
	public int count=Integer.MAX_VALUE-1;

	@Option(name="-labeled",usage="train a labeled model")
	public boolean labeled;


	@Option(name="-cores",aliases="-threads",usage="Number of threads available for threadpools (default half of available)")
	public static int cores = (int) Math.max(1, Math.floor(Runtime.getRuntime().availableProcessors()/2.0));

	@Option(name="-debug",usage="more debug output")
	public static boolean DEBUG=false;
	
	@Option(name="-lblFilter",usage="set label filter")
	public LBLFilter labelFilter=LBLFilter.None;

	
	public static enum TransitionSystem { 
		ArcS(new TR[]{TR.SH,TR.LA,TR.RAS}),
		ArcSLazy(new TR[]{TR.SH,TR.LA,TR.RAS}),
		ArcSOldSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSNewSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSOptimalSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSDynSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSFullDynSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSLASWDynSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSConLASWDynSwap(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSSwapDynDelay(new TR[]{TR.SH,TR.LA,TR.RAS,TR.SW}),
		ArcSDynDelay(new TR[]{TR.SH,TR.LA,TR.RAS}),
		; 
	
		public final TR[] trs;
		TransitionSystem(TR[] trs){
			this.trs=trs;
		}
	}
	
	public static enum UpdateType {
		BeamEarly, BeamMaxVio,
		BeamEarlyGoldBeam, BeamMaxVioGoldBeam,
		BeamDelayLaSO, BeamDelayLaSOGoldBeam,
		Greedy, GreedyDelayed, BeamMaxVioGoldBeamOpt
	}
	
	@Option(name="-upd",aliases="-updateType",usage="How the learning works")
	public UpdateType ut=UpdateType.BeamMaxVioGoldBeamOpt;
	
	
	public static enum Oracle {	Static,Dynamic,Model,File }
	
	@Option(name="-system",aliases="--transitionSystem",usage="What transition system to use")
	public TransitionSystem transitionSystem=TransitionSystem.ArcSNewSwap;
	@Option(name="-oracle")
	public Oracle oracle=Oracle.Static;
	
	@Option(name="-trFile")
	public File trSeqFile;

	@Option(name="-overRideBeam",aliases="-beamOverride")
	public int overRideBeam=-1;

	@Option(name="-mergeEqual",usage="Merge equal states during beam search")
	public boolean mergeEquivalentStates=false;
	
	@Option(name="-dotPrefix",aliases="-dot",usage="file prefix for dot files for dynamic oracle plotting")
	public String dotPrefix=null;
	
	@Option(name="-onlyNpAC",aliases="-arcs")
	public int keepNonProjArcCount=-1;

	@Option(name="-outTRFilePF")
	public static String outTRFilePF=null;
	
	@Option(name="-left")
	public int left=-1;
	
	@Option(name="-right")
	public int right=-1;
	
	@Option(name="-maxLength")
	public int maxLength=-1;
	
	@Option(name="-lf",aliases="-lossFunction",usage="loss function for learning")
	public static LossFunction.Type lossFunctionType=LossFunction.Type.SwapHammingLabelWeighted;
	
	@Option(name="-fs",aliases="-featureSet",usage="feature set")
	public FeatureSet fs=FeatureSet.Default;
	
	@Option(name="-format",usage="set input format")
	public Format format=Format.CoNLL06;
	
	@Option(name="-rtl",aliases="-reversed",usage="Pass this during training to do left to right decoding")
	public boolean reversed=false;
	
	//Formats for eval with differing gold and pred
	@Option(name="-goldFormat")
	public Format goldFormat=null;
	
	@Option(name="-predFormat")
	public Format predFormat=null;
	
	@Option(name="-useStag",aliases="-useStags",usage="use supertag features")
	public boolean useStag=false;
	
	@Option(name="-useStack",usage="use stacking features")
	public boolean useStack=false;
	
	@Option(name="-greedy",usage="train greedy way")
	public boolean trainGreedy=false;
	
	@Option(name="-goldDecodePolicy")
	public static GoldDecodePolicy goldDecodePolicy=GoldDecodePolicy.Beam;

	@Option(name="-goldModel")
	public static File goldModel=null;
	
	@Option(name="-goldBeam")
	public static int goldBeam=DEFAULT_BEAM_SIZE;

	@Option(name="-dynDelay",usage="delay using dynamic oracle until this iteratio")
	public static int dynDelayIter=5;
	
	@Option(name="-statSys")
	public static TransitionSystem statSystem=null;
	
	@Option(name="-dynSys")
	public static TransitionSystem dynSystem=null;
	
	@Option(name="-folds")
	public int folds=5;
	
	@Option(name="-cvShare")
	public boolean cvSharedSymTab=false;
	
	@Option(name="-octave")
	public static boolean octave=false;

	@Option(name="-docID")
	public String docID=null;
	
	@Option(name="-fss")
	public static int featureSinkSize=1500;
	
	@Option(name="-x0")
	public static boolean x0=false;
	@Option(name="-x1")
	public static boolean x1=false;
	@Option(name="-x2")
	public static boolean x2=false;
	@Option(name="-x3")
	public static boolean x3=false;
	@Option(name="-x4")
	public static boolean x4=false;
	@Option(name="-x5")
	public static boolean x5=false;
	@Option(name="-x6")
	public static boolean x6=false;
	@Option(name="-x7")
	public static boolean x7=false;
	@Option(name="-x8")
	public static boolean x8=false;
	@Option(name="-x9")
	public static boolean x9=false;
	
	
	public static enum ShuffleStrategy {
		Never(false,false), 
		Once(true,false), 
		Every(true,true);
		
		public boolean shuffleFirstIteration;
		public boolean shuffleBetweenIterations;
		ShuffleStrategy(boolean first,boolean between){
			this.shuffleFirstIteration=first;
			this.shuffleBetweenIterations=between;
		}
	}
	@Option(name="-shuffle",aliases="-shuffleStrategy",usage="when to shuffle")
	public static ShuffleStrategy shuffleStrategy=ShuffleStrategy.Every;

	@Option(name="-augLoss")
	public static boolean AUG_LOSS=false;

	@Option(name="-strict")
	public static boolean STRICT=true;
	
	public void sanityCheckTestEvery() {
		if(testEveryIter){
			boolean fail=false;
			if(test==null){
				fail=true;
				System.err.println("You can't test every iteration if you don't specify a test file (-test)");
			}
			if(gold==null){
				fail=true;
				System.err.println("You can't test every iteration if you don't specify a gold file (-gold)");
			}
			if(fail)
				System.exit(1);
		}
	}
	
	public void warnNoModel(){
		if(cvSharedSymTab && model==null){
			System.err.println("Need to specify model when doing cvShare -- exiting");
			System.exit(1);
		}
		if(model==null){
			System.out.println("\n\nNOTE: You did not specify a model file so the model won't be written to File. You might want to restart and pass a -model parameter.\n\n");
		}
	}
	
	
	public Options(String[] args){
		CmdLineParser clp=new CmdLineParser(this);
		try {
			clp.parseArgument(args);
		} catch (CmdLineException e) {
			e.printStackTrace();
			System.err.println("Failed to parse cmd line, aborting.");
			System.err.println(e.getMessage());
			System.err.println();
			System.err.println("Valid options are:");
			printUsage(System.err);
			System.exit(1);
		}
	}
	private Options(){}
	public static void printUsage(OutputStream out){
		CmdLineParser clp=new CmdLineParser(new Options());
		clp.printUsage(out);
	}	
	public static void main(String[] args){
		printUsage(System.out);
	}
	
	public void done(){
		DBO.println("done() at "+new Date());
		DBO.println("Total time "+Util.insertCommas(System.currentTimeMillis()-optionsCreate.getTime()));
		ThreadPoolSingleton.shutdown();
		DBO.close();
	}
	
	public String toString(){
		StringBuilder sb=new StringBuilder("OPTIONS:\n");
		sb.append(String.format("%20s %s\n","Train File:",this.train==null?"null":this.train.getAbsolutePath()));
		sb.append(String.format("%20s %s\n","Test File:",this.test==null?"null":this.test.getAbsolutePath()));
		sb.append(String.format("%20s %s\n","Out File:",this.out==null?"null":this.out.getAbsolutePath()));
		sb.append(String.format("%20s %s\n","Gold File:",this.gold==null?"null":this.gold.getAbsolutePath()));
		sb.append(String.format("%20s %s\n","Model File:",this.model==null?"null":this.model.getAbsolutePath()));
//		sb.append(String.format("%20s %s\n",null,null));
//		sb.append(String.format("%20s %s\n",null,null));
//		sb.append(String.format("%20s %s\n",null,null));
		return sb.toString();
	}
	
}
