package ims.trans;

import java.util.LinkedList;
import java.util.List;

public class PushbackBuffer implements IBuffer {

	private final int nxt;
	private final int last;
	private final LinkedList<State> pushedBack;
	
	public PushbackBuffer(int nxt, int last, State s) {
		this.nxt=nxt;
		this.last=last;
		pushedBack=new LinkedList<State>();
		pushedBack.add(s);
	}
	
	private PushbackBuffer(PushbackBuffer old){
		this.nxt=old.nxt;
		this.last=old.last;
		this.pushedBack=new LinkedList<State>(old.pushedBack);
	}
	
	private PushbackBuffer(PushbackBuffer old,State s){
		this(old);
		this.pushedBack.addFirst(s);
	}

	public PushbackBuffer(int nxt, int last, List<State> subList) {
		this.nxt=nxt;
		this.last=last;
		this.pushedBack=new LinkedList<State>(subList);
	}

	@Override
	public IBuffer pushBack(State s) {
		return new PushbackBuffer(this,s);
	}
	
	@Override
	public int size() {
		return last-nxt+1+pushedBack.size();
	}
	
	@Override
	public int get(int n) {
		if(n<pushedBack.size())
			return pushedBack.get(n).s0;
		
		int r=nxt+(n-pushedBack.size());
		if(r>=last)
			return -1;
		else
			return r;		
	}
	
	@Override
	public IBuffer advance() {
		if(pushedBack.size()==1)
			return SimpleBuffer.getBuffer(nxt, last);
		else
			return new PushbackBuffer(nxt,last,pushedBack.subList(1, pushedBack.size()));
	}
	
	@Override
	public State peekPushedBack() {
		if(pushedBack.isEmpty())
			return null;
		else
			return pushedBack.getFirst();
	}

	@Override
	public boolean hasSwappedItems() {
		return !pushedBack.isEmpty();
	}

	@Override
	public boolean equals(IBuffer b) {
		return b.equals(this);
	}

	@Override
	public boolean equals(SimpleBuffer other) {
		if(size()!=other.size())
			return false;
		int max=this.pushedBack.size()+1;
		int min=Math.min(max, this.size());
		for(int i=0;i<min;++i)
			if(this.get(i)!=other.get(i))
				return false;
		return true;
	}

	@Override
	public boolean equals(PushbackBuffer other) {
		if(size()!=other.size())
			return false;
		int max=Math.max(this.pushedBack.size(),other.pushedBack.size())+1;
		int min=Math.min(max, this.size());
		for(int i=0;i<min;++i)
			if(this.get(i)!=other.get(i))
				return false;
		return true;
	}

	@Override
	public IBuffer deepCopy() {
		return new PushbackBuffer(this);
	}

	@Override
	public int getNoPushedBack() {
		return pushedBack.size();
	}
	
	public String toString(){
		StringBuilder sb=new StringBuilder("[");
		for(State s:pushedBack)
			sb.append(s.s0).append(",");
		sb.append(nxt).append("..").append(last);
		return sb.append("]").toString();
	}
	
}
