package ims.trans;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ims.data.ISentence;
import ims.data.Stack;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.oracle.IConfiguration;
import ims.oracle.Configuration.TR;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.TransitionFactory.LabeledAbstractTransition;
import ims.util.MinHeap.HasScore;
import ims.util.IntPair;
import ims.util.Pair;

public class State implements IConfiguration,HasScore {	
		
	public static final int NOHEAD_LABEL_L=AbstractSymbolMapping.NONE_IDX;
	public static final int UNKNOWN_LABEL_L=AbstractSymbolMapping.UNK_IDX;
	public static final short[] EMPTY_SHORT_ARRAY=new short[0];
	
	public final State tail;  //Stack
	public final IBuffer buf; //Buffer
	
	//To recover a tree given a final state, follow these back pointers and do the actions
	public final ITransition lastAction;
	public final State previous;
	
	public final double totalScore;
	public final int totalTransitions;
	
	//For training
	public boolean correct; //Default, *must* reassign this during training

	//For features
	public final int s0,s0L,s0R,s0L2,s0R2,s0LR,s0RL;
	
//	private final String trSignature; //Signature in terms of transitions
	protected final int[] treeSig;
	protected final int[] lblSig;
	
	//Track remaining deps (for dynamic oracle training)
	private final short[] remainingDepsLeft;
	private final short[] remainingDepsRight;
	
	//Track the number of left and right deps on every token (for features)
	private final short[] leftDeps;
	private final short[] rightDeps;
	
	//Track arc labels on left and right side
	public final short[][] leftLabels;
	public final short[][] rightLabels; 

	public final int sentenceLength;
	
	public State deepCopy(){
		if(true)
			throw new Error("need to fix something here");
		IBuffer buf=this.buf.deepCopy();
		short[] remainingLeft=this.remainingDepsLeft==null?null:Arrays.copyOf(this.remainingDepsLeft,this.remainingDepsLeft.length);
		short[] remainingRight=this.remainingDepsRight==null?null:Arrays.copyOf(this.remainingDepsRight,this.remainingDepsRight.length);
		short[] leftDeps=Arrays.copyOf(this.leftDeps, this.leftDeps.length);
		short[] rightDeps=Arrays.copyOf(this.rightDeps, this.rightDeps.length);

		int[] treeSig=deepCopy(this.treeSig);
		int[] lblSig=deepCopy(this.lblSig);

		//Need to make proper copy here
		short[][] leftLabels=this.leftLabels;
		short[][] rightLabels=this.rightLabels;
		return new State(s0,s0L,s0R,tail,lastAction,previous,totalScore,correct,buf,totalTransitions,remainingLeft,remainingRight,treeSig,lblSig,leftDeps,rightDeps,leftLabels,rightLabels,s0L2,s0R2,s0LR,s0RL,sentenceLength);
	}
	
	protected State(int s0, int s0l, int s0r, State tail, ITransition lastAction, State previous, double totalScore,boolean correct,IBuffer buf,int totalTransitions,short[] remainingDepsLeft,short[] remainingDepsRight,int[] treeSig,int[] lblSig,short[] leftDeps,short[] rightDeps,short[][] leftLabels,short[][] rightLabels,int s0L2,int s0R2,int s0LR,int s0RL,int senLength) {
		if(buf==null)
			System.err.println("here");
		this.s0 = s0;
		s0L = s0l;
		s0R = s0r;
		this.tail = tail;
		this.lastAction = lastAction;
		this.previous = previous;
		this.correct=correct;
		this.totalScore=totalScore;
		this.totalTransitions=totalTransitions;
		this.buf=buf;
		if(s0==-1 && tail!=null)
			throw new Error("!");
		this.treeSig=treeSig;
		this.lblSig=lblSig;
		this.remainingDepsLeft=remainingDepsLeft;
		this.remainingDepsRight=remainingDepsRight;
		this.leftDeps=leftDeps;
		this.rightDeps=rightDeps;
		this.leftLabels=leftLabels;
		this.rightLabels=rightLabels;
		this.s0L2=s0L2;
		this.s0R2=s0R2;
		this.s0LR=s0LR;
		this.s0RL=s0RL;
		this.sentenceLength=senLength;
	}
	
	public static State getInitialState(int length,boolean labeled,int[] heads){
		final int[] treeSig=new int[length];
		Arrays.fill(treeSig, -1);
		treeSig[0]=ISentence.ROOT_HEAD;;
		final int[] lblSig;
		if(labeled){
			lblSig=new int[length];
			Arrays.fill(lblSig, -1);
			lblSig[0]=AbstractSymbolMapping.INIT_IDX; //This is what we will have in Instance.lbls[0], since the root label is the first one that gets registered in the symbol table
		} else {
			lblSig=null;
		}
		IBuffer buf=SimpleBuffer.getBuffer(1, length-1);
		short[] left=null;
		short[] right=null;
		if(heads!=null){
			left=new short[heads.length];
			right=new short[heads.length];
			for(int i=1;i<heads.length;++i){
				if(heads[i]<i) //head on left, so it's a right dependent of heads[i]
					++right[heads[i]];
				else
					++left[heads[i]];
			}
		}
		short[][] leftLabels;
		short[][] rightLabels;
		if(labeled){
			leftLabels=new short[length][];
			rightLabels=new short[length][];
			Arrays.fill(leftLabels, EMPTY_SHORT_ARRAY);
			Arrays.fill(rightLabels, EMPTY_SHORT_ARRAY);
		} else {
			leftLabels=null;
			rightLabels=null;
		}
		State s=new State(0,-1,-1,null,null,null,0,true,buf,0,left,right,treeSig,lblSig,new short[length],new short[length],leftLabels,rightLabels,-1,-1,-1,-1,length);
		return s;
	}
	
	public String toString(){
		StringBuilder sb=new StringBuilder("Correct: "+correct+" b0: "+buf.get(0)+" s0: "+s0);
		sb.append("\nScore: "+score());
		if(lastAction==null)
			return sb.toString();
		sb.append(" tr: ").append(lastAction.getFullTransitionLabel()).append("\n");
		sb.append("signature: ").append(getTransitionSeqSignature()).append('\n');
		int transCount=0;
		List<String> allTrans=new ArrayList<String>();
		allTrans.add(lastAction.getFullTransitionLabel());
		for(State p=this.previous;p!=null;p=p.previous){
			transCount++;
			if(p.lastAction!=null)
				allTrans.add(p.lastAction.getFullTransitionLabel());
		}
			
		Collections.reverse(allTrans);
		for(String s:allTrans)
			sb.append(s).append(' ');
		sb.append('\n').append(transCount);
		sb.append('\n').append("Stack rev: ");
		for(State x=this;x!=null;x=x.tail)
			sb.append(x.s0).append(",");
		sb.append('\n').append("Buf: ").append(buf.toString());
		return sb.toString();
	}

	public static ITransition[] getTransitionSequence(State s) {
		ITransition[] seq=new ITransition[s.totalTransitions];
		State cur=s;
		for(int i=seq.length-1;i>=0;--i){
			seq[i]=cur.lastAction;
			cur=cur.previous;
		}
		return seq;
	}

	@Override
	public double score() {
		return totalScore;
//		return getAvgScore();
	}
	
//	private double getAvgScore(){
//		if(totalTransitions==0)
//			return 0.d;
//		else
//			return totalScore/totalTransitions;
//	}
	
	public int hashCode(){
		return 3*s0+7*buf.get(0);
	}
	
	public boolean equals(Object other){
		return equals((State) other);
	}
	
	public boolean equals(State other){
		if(!buf.equals(other.buf))   //Equal buffers
			return false;
		if(!equivalentStacks(other)) //Equal stacks
			return false;
		
		return Arrays.equals(treeSig, other.treeSig) && Arrays.equals(lblSig, other.lblSig); //Equal trees
	}
	
	private boolean equivalentStacks(State other) {
		if(s0!=other.s0)
			return false;
		if(tail==null){
			if(other.tail==null)
				return true;
			else
				return false;
		} else if(other.tail==null){
			return false;
		} else { //Else neither are null, so we recurse
			return tail.equivalentStacks(other.tail);
		}
	}

	public boolean isTerminal(){
		return buf.get(0)==-1 && tail==null;
	}
	
	public State doShift(ITransition tr,double newScore){
		State tail=this;
		int s0=this.buf.get(0);
		if(s0==-1){
			System.err.println("here ");
			System.err.println(this.buf.get(0));
			throw new Error("!");
		}
		int s0l=-1;
		int s0r=-1;
		int s0l2=-1;
		int s0r2=-1;
		int s0lr=-1;
		int s0rl=-1;
		State pushedBack=this.buf.peekPushedBack();
		if(pushedBack!=null){
			if(s0!=pushedBack.s0)
				throw new Error("!"); //shouldn't happen
			s0l=pushedBack.s0L;
			s0r=pushedBack.s0R;
			s0l2=pushedBack.s0L2;
			s0r2=pushedBack.s0R2;
			s0lr=pushedBack.s0LR;
			s0rl=pushedBack.s0RL;
		}
		IBuffer buf=this.buf.advance();
		int[] treeSig=this.treeSig;
		int[] lblSig=this.lblSig;
		short[] remainingLeft=this.remainingDepsLeft;
		short[] remainingRight=this.remainingDepsRight;
		short[] lD=this.leftDeps;
		short[] rD=this.rightDeps;
		return new State(s0,s0l,s0r,tail,tr,this,newScore,this.correct,buf,this.totalTransitions+1,remainingLeft,remainingRight,treeSig,lblSig,lD,rD,this.leftLabels,this.rightLabels,s0l2,s0r2,s0lr,s0rl,sentenceLength);
	}
	
	public State doSwap(ITransition tr,double newScore){
		State s1=this.tail; //This is the guy that gets pushed back on the stack
		int s0=this.s0;
		int s0r=this.s0R;
		int s0l=this.s0L;
		int s0l2=this.s0L2;
		int s0r2=this.s0R2;
		int s0lr=this.s0LR;
		int s0rl=this.s0RL;
		State tail=s1.tail;  //Reset the tail
		IBuffer buf=this.buf.pushBack(s1);
		int[] treeSig=this.treeSig;
		int[] lblSig=this.lblSig;
		short[] remainingLeft=this.remainingDepsLeft;
		short[] remainingRight=this.remainingDepsRight;
		short[] lD=this.leftDeps;
		short[] rD=this.rightDeps;
		return new State(s0,s0l,s0r,tail,tr,this,newScore,this.correct,buf,this.totalTransitions+1,remainingLeft,remainingRight,treeSig,lblSig,lD,rD,this.leftLabels,this.rightLabels,s0l2,s0r2,s0lr,s0rl,sentenceLength);
	}
	
	public State doSB(ITransition tr,double newScore){
		int s0=this.s0;
		int s0r=this.s0R;
		int s0l=this.s0L;
		int s0l2=this.s0L2;
		int s0r2=this.s0R2;
		int s0lr=this.s0LR;
		int s0rl=this.s0RL;
		State tail=this.tail;
		IBuffer buf=this.buf;
		int[] treeSig=this.treeSig;
		int[] lblSig=this.lblSig;
		short[] remainingLeft=this.remainingDepsLeft;
		short[] remainingRight=this.remainingDepsRight;		
		short[] lD=this.leftDeps;
		short[] rD=this.rightDeps;
		return new State(s0,s0l,s0r,tail,tr,this,newScore,this.correct,buf,this.totalTransitions+1,remainingLeft,remainingRight,treeSig,lblSig,lD,rD,this.leftLabels,this.rightLabels,s0l2,s0r2,s0lr,s0rl,sentenceLength);
	}

	public State doLeftArc(LabeledAbstractTransition tr,double newScore){
		int s0=this.s0; //Head
		int s0l=this.tail.s0; //Dep
		int head=this.s0;
		int dep=this.tail.s0;
		int s0r=this.s0R;
		int s0l2=this.tail.s0L;
		int s0r2=this.s0R2;
		int s0lr=this.tail.s0R;
		int s0rl=this.s0RL;
		State tail=this.tail.tail;
		IBuffer buf=this.buf;
		int[] treeSig=deepCopy(this.treeSig);
		treeSig[dep]=head;
		final int[] lblSig;
		if(this.lblSig!=null){
			lblSig=deepCopy(this.lblSig);
			lblSig[dep]=tr.l;
		} else {
			lblSig=null;
		}
		short[][] leftLabels;
		short[][] rightLabels;
		if(this.leftLabels!=null){ //Then we have to do something here
			if(dep<head){ //left dependent
				leftLabels=cloneAndAdd(this.leftLabels,head,(short) tr.l);
				rightLabels=this.rightLabels;
			} else { //right dependent
				leftLabels=this.leftLabels;
				rightLabels=cloneAndAdd(this.rightLabels,head,(short) tr.l);
			}
		} else {
			leftLabels=null;
			rightLabels=null;
		}
		
		short[] remainingLeft;
		short[] remainingRight;
		
		short[] leftDeps;
		short[] rightDeps;

		
		if(dep<head){
			rightDeps=this.rightDeps;
			remainingRight=this.remainingDepsRight;
			leftDeps=Arrays.copyOf(this.leftDeps, this.leftDeps.length);
			leftDeps[head]++;

			if(this.remainingDepsLeft!=null){
				remainingLeft=Arrays.copyOf(this.remainingDepsLeft, this.remainingDepsLeft.length);
				remainingLeft[head]--;
			} else {
				remainingLeft=null;
			}
		} else {
			leftDeps=this.leftDeps;
			remainingLeft=this.remainingDepsLeft;
			rightDeps=Arrays.copyOf(this.rightDeps, this.rightDeps.length);
			rightDeps[head]++;
			if(remainingDepsRight!=null){
				remainingRight=Arrays.copyOf(this.remainingDepsRight, this.remainingDepsRight.length);
				remainingRight[head]--;
			} else {
				remainingRight=null;
			}
		}

		State r=new State(s0,s0l,s0r,tail,tr,this,newScore,this.correct,buf,this.totalTransitions+1,remainingLeft,remainingRight,treeSig,lblSig,leftDeps,rightDeps,leftLabels,rightLabels,s0l2,s0r2,s0lr,s0rl,sentenceLength);
		return r;
	}
	
	public State doRightArcS(LabeledAbstractTransition tr,double newScore){
		int s0=this.tail.s0;
		int s0l=this.tail.s0L;
		int s0r=this.s0;
		int s0l2=this.tail.s0L2;
		int s0r2=this.s0R;
		int s0lr=this.s0LR;
		int s0rl=this.s0L;
		State tail=this.tail.tail;
		int head=this.tail.s0;
		int dep=this.s0;
		IBuffer buf=this.buf;
		int[] treeSig=deepCopy(this.treeSig);
		treeSig[dep]=head;
		final int[] lblSig;
		if(this.lblSig!=null){
			lblSig=deepCopy(this.lblSig);
			lblSig[dep]=tr.l;
		} else {
			lblSig=null;
		}

		short[][] leftLabels;
		short[][] rightLabels;
		if(this.leftLabels!=null){ //Then we have to do something here
			if(dep<head){ //left dependent
				leftLabels=cloneAndAdd(this.leftLabels,head,(short) tr.l);
				rightLabels=this.rightLabels;
			} else { //right dependent
				leftLabels=this.leftLabels;
				rightLabels=cloneAndAdd(this.rightLabels,head,(short) tr.l);
			}
		} else {
			leftLabels=null;
			rightLabels=null;
		}

		short[] leftDeps;
		short[] rightDeps;

		short[] remainingLeft;
		short[] remainingRight;
		
		if(dep<head){ //it's a left dependent
			rightDeps=this.rightDeps;
			remainingRight=this.remainingDepsRight;
			leftDeps=Arrays.copyOf(this.leftDeps, this.leftDeps.length);
			leftDeps[head]++;
			if(this.remainingDepsLeft!=null){
				remainingLeft=Arrays.copyOf(this.remainingDepsLeft, this.remainingDepsLeft.length);
				remainingLeft[head]--;
			} else {
				remainingLeft=null;
			}
		} else {
			leftDeps=this.leftDeps;
			remainingLeft=this.remainingDepsLeft;
			rightDeps=Arrays.copyOf(this.rightDeps, this.rightDeps.length);
			rightDeps[head]++;
			if(this.remainingDepsRight!=null){
				remainingRight=Arrays.copyOf(this.remainingDepsRight, this.remainingDepsRight.length);
				remainingRight[head]--;
			} else {
				remainingRight=null;
			}
		}

		State r=new State(s0,s0l,s0r,tail,tr,this,newScore,this.correct,buf,this.totalTransitions+1,remainingLeft,remainingRight,treeSig,lblSig,leftDeps,rightDeps,leftLabels,rightLabels,s0l2,s0r2,s0lr,s0rl,sentenceLength);
		return r;
	}
	
	private static short[][] cloneAndAdd(short[][] a,int head,short lbl){
		short[][] r=new short[a.length][];
		for(int i=0;i<a.length;++i)
			if(i==head)
				continue;
			else
				r[i]=a[i];
		short[] b=new short[a[head].length+1];
		System.arraycopy(a[head], 0, b, 0, a[head].length);
		b[a[head].length]=lbl;
		r[head]=b;
		return r;
	}
	
	
	public static String getULBLSig(IConfiguration c){
		StringBuilder sb=new StringBuilder();
		c.appendStackSig(sb);
		sb.append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=0,m=c.bufferSize();i<m;++i)
			sb.append((char) c.peekBuffer(i));
		sb.append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=0,m=c.sentenceSize();i<m;++i)
			sb.append((char) c.getHead(i));
		return sb.toString();
	}
	
	public static String getLBLSig(IConfiguration c){
		StringBuilder sb=new StringBuilder();
		c.appendStackSig(sb);
		sb.append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=0,m=c.bufferSize();i<m;++i)
			sb.append((char) c.peekBuffer(i));
		sb.append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=0,m=c.sentenceSize();i<m;++i)
			sb.append((char) c.getHead(i));
		sb.append(IConfiguration.STATE_SIG_SEPARATOR);
		for(int i=0,m=c.sentenceSize();i<m;++i)
			sb.append((char) c.getLbl(i));
		return sb.toString();
	}

	public StringBuilder appendStackSig(StringBuilder sb) {
		if(tail!=null) //Bottom up, descend before picking up this one
			tail.appendStackSig(sb);
		sb.append((char) s0);
		return sb;
	}

	public int getHead(int dep) {
		return treeSig[dep];
	}
	
	@Override
	public int bufferSize() {
		return buf.size();
	}

	@Override
	public int peekBuffer(int d) {
		return buf.get(d);
	}

	@Override
	public int peekStack(int d) {
		State t=this;
		for(int p=0;p<d;++p){
			if(t.tail==null)
				return -1;
			t=t.tail;
		}
		return t.s0;
	}

	@Override
	public boolean hasHead(int dep) {
		return getHead(dep)>=0;
	}

	@Override
	public boolean hasMinTwoStackElements() {
		return this.tail!=null;
	}
	
	@Override
	public boolean hasAllDeps(int head) {
		return hasAllLeftDeps(head) && hasAllRightDeps(head);
	}
	
	@Override
	public boolean hasAllLeftDeps(int head) {
		return remainingDepsLeft[head]==0;
	}

	@Override
	public boolean hasAllRightDeps(int head) {
		return remainingDepsRight[head]==0;
	}

	public int getLabel(int dep) {
		if(lblSig==null){
			if(getHead(dep)>=0){
				return AbstractSymbolMapping.NONE_IDX;
			} else {
				return AbstractSymbolMapping.UNK_IDX;
			}
		}
		int l=lblSig[dep];
		if(l>0)
			return l;
		else
			return AbstractSymbolMapping.UNK_IDX;
	}
	
	public Pair<int[], String[]> getPredTree(String[] deprelLookup) {
		int[] heads=new int[sentenceLength];
		String[] labels=new String[sentenceLength];
		for(int i=0;i<sentenceLength;++i){
			heads[i]=treeSig[i];
			if(lblSig!=null){
				int l=lblSig[i];
				labels[i]=l>=0?deprelLookup[l]:"UNDEF";
			}
		}
		return new Pair<int[],String[]>(heads,labels);
	}

	public int getLeftDeps(int head){
		return leftDeps[head];
	}
	
	public int getRightDeps(int head){
		return rightDeps[head];
	}

	@Override
	public int stackSize() {
		throw new Error("not implemented");
	}

	@Override
	public TokenState getTokenState(int i) {
		throw new Error("not implemented (yet)");
	}

	@Override
	public int sentenceSize() {
		return sentenceLength;
	}

	@Override
	public int getBufferNoPushedBack() {
		return buf.getNoPushedBack();
	}

	@Override
	public String getTransitionSeqSignature() {
		StringBuilder sb=new StringBuilder();
		sb.append('0');
		for(ITransition t:getTransitionSequence(this))
			sb.append(t.getTR().k);
		return sb.toString();
	}

	@Override
	public int getLbl(int i) {
		return lblSig[i];
	}

	@Override
	public String getULBLStateSig() {
		return getULBLSig(this);
	}

	@Override
	public String getLBLStateSig() {
		return getLBLSig(this);
	}

	@Override
	public int getRemainingDeps(int head) {
		return remainingDepsLeft[head]+remainingDepsRight[head];
	}

	@Override
	public Stack getStackCopy() {
		Stack stack=new Stack();
		addStackElementsRecursive(stack);
		return stack;
	}

	private void addStackElementsRecursive(Stack stack) {
		if(tail!=null)
			tail.addStackElementsRecursive(stack);
		if(s0>0)
			stack.add(s0);
	}
	
	public int countSwaps(){
		int i=0;
		State c=this;
		while(c.lastAction!=null){
			if(c.lastAction.getTR()==TR.SW)
				++i;
			c=c.previous;
		}
		return i;
	}

	public int getTotalAssignedHeads() {
		int c=0;
		for(int i=1;i<sentenceLength;++i)
			if(hasHead(i))
				++c;
		return c;
	}
	
	
	private static int[] deepCopy(int[] intArr) {
		if(intArr==null)
			return null;
		return Arrays.copyOf(intArr, intArr.length);
	}
	
	public boolean equalTRSeq(State other){
		if(this.totalTransitions!=other.totalTransitions)
			return false;
		//Could do recursion here, but iteration feels better.
		State thisCur=this;
		State otherCur=other;
		for(int i=this.totalTransitions;i>0;--i){
			if(thisCur.lastAction!=otherCur.lastAction)
				return false;
			thisCur=thisCur.previous;
			otherCur=otherCur.previous;
		}
		return true;
	}
	
	public List<String> recoverTRSeqStringList(){
		List<String> r=new ArrayList<String>(totalTransitions);
		recoverTRSeqStringListRecursive(r);
		return r;
	}
	
	private void recoverTRSeqStringListRecursive(List<String> l){
		if(this.previous!=null){
			this.previous.recoverTRSeqStringListRecursive(l);
			l.add(this.lastAction.getFullTransitionLabel());
		}
	}

	@Override
	public int getTotalTransitions() {
		return totalTransitions;
	}
	

	/*
	 * Returns a set of IntPairs (s0,s1) for pairs that have been swapped in the history
	 */
	public Set<IntPair> getSwappedPairsInTRHistory() {
		Set<IntPair> r=new HashSet<IntPair>();
		State cur=this;
		while(cur.lastAction!=null){
			if(cur.lastAction.getTR()==TR.SW){
				cur=cur.previous;
				int s0=cur.s0;
				int s1=cur.tail.s0;
				r.add(new IntPair(s0,s1));
			} else {
				cur=cur.previous;
			}
		}
		return r;
	}

	public List<Integer> getSentenceBoundaries() {
		List<Integer> sentenceInitials=new ArrayList<>();
		State cur=this;
		while(cur.lastAction!=null){
			if(cur.lastAction.getTR()==TR.SB)
				sentenceInitials.add(cur.peekBuffer(0));
			cur=cur.previous;
		}
		Collections.reverse(sentenceInitials);
		return sentenceInitials;
	}
	
}
