package ims.trans;

import java.io.IOException;

import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.trans.util.Eval;
import ims.util.Accuracy;
import ims.util.DBO;
import ims.util.Pair;
import ims.util.ZipUtil;

public class Test {

	public static void main(String[] args) throws IOException, ClassNotFoundException{
		Options options=new Options(args);
		DBO.println("Loading model");
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		DBO.println("Done.");
		Decoder d=new Decoder(m);
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.test, options.format.sentenceClass,m.reversed);
		SentenceWriter writer=ReaderWriterFactory.getWriter(options.out,m.reversed);
		int beam=m.beamSize;
		if(options.overRideBeam>0)
			beam=options.overRideBeam;
		DBO.println("Parsing");
		test(d, reader, writer, beam);
		DBO.println("Done");
		if(options.gold!=null){
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(options.out, options.gold, options.format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		}
		writer.close();
		options.done();
	}

	private static void test(Decoder d, SentenceReader<?> reader, SentenceWriter writer,int beam)throws IOException {
		int sentCount=0;
		int erase=0;
		DBO.printNoPrefix("Sent: ");
		long start=System.currentTimeMillis();
		for(ISentence s:reader){
			State state=d.parse(s,beam);
			Pair<int[],String[]> p=d.state2tree(state);
			int[] heads=p.getLeft();
			String[] lbl=p.getRight();
			s.setPredHeads(heads);
			s.setPredLabels(lbl);
			writer.write(s);
			sentCount++;
			if(sentCount%111==0 || Options.DEBUG)
				erase=DBO.eraseAndPrint(erase, getInfoString(sentCount,System.currentTimeMillis()-start));
		}
		erase=DBO.eraseAndPrint(erase, getInfoString(sentCount,System.currentTimeMillis()-start));
		DBO.println();
	}

	private static String getInfoString(int sentCount, long l) {
		long tpi=l/sentCount;
		return String.format("%-5d   (%d/s)", sentCount,tpi);
	}

	public static void applyDecoder(Decoder d, SentenceReader<?> reader, SentenceWriter writer,int beam) throws IOException {
		test(d,reader,writer,beam);
	}
	
}
