package ims.trans.doc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import ims.data.Document;
import ims.data.ISentence;
import ims.data.Instance;
import ims.io.DocumentWriter;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.trans.Decoder;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.State;
import ims.trans.util.Eval;
import ims.util.Accuracy;
import ims.util.DBO;
import ims.util.Pair;
import ims.util.ZipUtil;

public class Test {

	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		DBO.println("Loading model");
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		DBO.println("Done.");
		
		testModel(m,options);
		
		if(options.gold!=null){
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(options.out, options.gold, options.format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		}
		
		options.done();
	}
	
	
	private static void testModel(Model m,Options options) throws IOException{
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.test, options.format);
		DocumentWriter    writer=new DocumentWriter(options.out);
		testModel(m,reader,writer,m.beamSize);
		writer.close();
	}
	
	public static <T extends ISentence> void testModel(Model m,SentenceReader<T> reader,DocumentWriter writer,int beamSize) throws IOException{
		Decoder decoder=new Decoder(m);
		long start=System.currentTimeMillis();
		int erase=0;
		int sentCount=0;
		DBO.printNoPrefix("Sent: ");
		for(T sent:reader){
			Instance inst=m.instanceMapping.createTestInstance(sent, m.symTab);
//			Instance inst=m.instanceMapping.createTestInstance(d,m.symTab);
			State s;
			if(beamSize>1)
				s=decoder.decodeBeam(inst, beamSize);
			else
				s=decoder.decodeGreedy(inst);
			
			
			
			Pair<int[],String[]> tree=decoder.state2tree(s);
			List<Integer> sentBoundaries=s.getSentenceBoundaries();
			sent.setPredHeads(tree.getLeft());
			sent.setPredLabels(tree.getRight());

			List<T> outSents=new ArrayList<>();
			List<String> sIDs=new ArrayList<>();

			int sId=1;
			int b=1;
			int tokOffset=0;
			for(Integer e:sentBoundaries){
				T p=(T) sent.getSubSentence(b,e,tokOffset);
				b=e;
				outSents.add(p);
				sIDs.add(Integer.toString(sId++));
				tokOffset+=p.getLength()-1;
			}
			//Then we need to get the last sentence too
			T p=(T)sent.getSubSentence(b,sent.getLength(),tokOffset);
			outSents.add(p);
			sIDs.add(Integer.toString(sId++));
			
			String dId=sent.getDocID();
			
			Document<T> d=new Document<T>(outSents, sIDs, dId);
			writer.write(d);
			++sentCount;
			erase=DBO.eraseAndPrint(erase, String.format("%-5d   (%d/s)", sentCount,(System.currentTimeMillis()-start)/sentCount));
		}
	}
	
}
