package ims.trans.transitions;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.State;

public class DefaultArcStandardSwapTransitionFilter implements ITransitionFilter {
	private static final long serialVersionUID = -4429857234458386815L;

	final int[] SH;
	final int[] SH_SW_LA_RA;
	final int[] SW_LA_RA;
	final int[] SH_LA_RA;
	final int[] LA_RA;
	
	final int[] SH_RA;
	final int[] RA;
	
	public DefaultArcStandardSwapTransitionFilter(SymbolTable symTab) {
		int leftArcs=0;
		int rightArcs=0;
		
		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition tr=symTab.transition2label[i];
			switch(tr.getTR()){
			case RAS: ++rightArcs; break;
			case LA:  ++leftArcs;  break;
			default:  //do nothing
			}
		}

		SH=new int[1];
		SH_SW_LA_RA=new int[2+leftArcs+rightArcs];            int a=0;
		SW_LA_RA=new int[1+leftArcs+rightArcs];               int b=0;
		SH_LA_RA=new int[1+leftArcs+rightArcs];               int c=0;
		LA_RA=new int[leftArcs+rightArcs];                    int d=0;
		
		SH_RA=new int[1+rightArcs];                           int g=0;
		RA=new int[rightArcs];                                int h=0;

		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition tr=symTab.transition2label[i];
			TR t=tr.getTR();
			int id=tr.getId();
			if(t!=TR.SB)
				SH_SW_LA_RA[a++]=id;
			if(t==TR.SH){
				SH[0]=id;
				SH_LA_RA[c++]=id;
				SH_RA[g++]=id;
			} else if(t==TR.SW){
				SW_LA_RA[b++]=id;
			} else if(t==TR.RAS){
				SW_LA_RA[b++]=id;
				SH_LA_RA[c++]=id;
				LA_RA[d++]=id;
				SH_RA[g++]=id;
				RA[h++]=id;
			} else if(t==TR.LA){
				SW_LA_RA[b++]=id;
				SH_LA_RA[c++]=id;
				LA_RA[d++]=id;
			}
		}
	}

	@Override
	public int[] getAllowedTransitions(State s, Instance inst) {
		boolean twoItemsOnStack=s.tail!=null; //two or more items on stack?
		
		if(!twoItemsOnStack)
			return SH;
		boolean canLeftArc=s.tail.s0!=0;		
		
		boolean canShift=s.buf.get(0)>0;
		boolean canSwap=twoItemsOnStack && s.tail.s0 < s.s0 && s.tail.s0>0; 
		
		if(canLeftArc){
			if(canShift && canSwap)
				return SH_SW_LA_RA;
			if(canShift)
				return SH_LA_RA;
			if(canSwap)
				return SW_LA_RA;
			return LA_RA;
		} else {
			if(canShift)
				return SH_RA;
			return RA;
		}
		
		//We can never have canSwap==true and canArc==false
	}

	@Override
	public boolean isPOSFiltered(){
		return false;
	}
}
