package ims.trans.transitions;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.State;
import ims.util.IntPair;
import ims.util.MutableInt;

public class POSArcStandardSwapTransitionFilter extends AbstractPOSTransitionFilter {
	private static final long serialVersionUID = -2763244261422770239L;


	final int[] SH;
	
	final int[][][] SH_SW_LA_RA;
	final int[][][] SW_LA_RA;
	final int[][][] SH_LA_RA;
	final int[][][] LA_RA;
	final int[][][] SH_RA;
	final int[][][] RA;
	
	final int[] d_SH_SW_LA_RA;
	final int[] d_SW_LA_RA;
	final int[] d_SH_LA_RA;
	final int[] d_LA_RA;
	final int[] d_SH_RA;
	final int[] d_RA;
	
	public POSArcStandardSwapTransitionFilter(SymbolTable symTab,List<Instance> instances){
		int sh=extractTR(TR.SH,symTab);
		this.SH=new int[]{sh};
		int sw=extractTR(TR.SW,symTab);
		
		//Now loop over instances a do counts and records
		Map<Integer,MutableInt> LAcounts=new TreeMap<Integer,MutableInt>();
		Map<Integer,MutableInt> RAcounts=new TreeMap<Integer,MutableInt>();
		
		Map<IntPair,Set<Integer>> seenLA=new HashMap<IntPair,Set<Integer>>();
		Map<IntPair,Set<Integer>> seenRA=new HashMap<IntPair,Set<Integer>>();
		
		//Populate the maps
		int maxPos = countArcs(symTab, instances, LAcounts, RAcounts, seenLA, seenRA);

		final int defaultLA=MutableInt.getMostFrequent(LAcounts);
		final int defaultRA=MutableInt.getMostFrequent(RAcounts);

		//Assign defaults
		d_SH_SW_LA_RA=new int[]{sh,sw,defaultLA,defaultRA};
		d_SW_LA_RA   =new int[]{sw,defaultLA,defaultRA};
		d_SH_LA_RA   =new int[]{sh,defaultLA,defaultRA};
		d_LA_RA      =new int[]{defaultLA,defaultRA};
		d_SH_RA      =new int[]{sh,defaultRA};
		d_RA         =new int[]{defaultRA};
		
		//Collect all head/dep pairs
		Set<IntPair> hdPairs=new TreeSet<IntPair>();
		hdPairs.addAll(seenLA.keySet());
		hdPairs.addAll(seenRA.keySet());

		SH_SW_LA_RA=new int[maxPos+1][maxPos+1][];
		SW_LA_RA   =new int[maxPos+1][maxPos+1][];
		SH_LA_RA   =new int[maxPos+1][maxPos+1][];
		LA_RA      =new int[maxPos+1][maxPos+1][];
		SH_RA      =new int[maxPos+1][maxPos+1][];
		RA         =new int[maxPos+1][maxPos+1][];

		for(IntPair hdp:hdPairs){
			int s0Pos=hdp.i1;
			int s1Pos=hdp.i2;

			Set<Integer> la=seenLA.get(hdp);
			Set<Integer> ra=seenRA.get(hdp);
			if(la==null)
				la=Collections.singleton(defaultLA);
			if(ra==null)
				ra=Collections.singleton(defaultRA);

			Set<Integer> union=new TreeSet<Integer>();
			union.addAll(la);
			union.addAll(ra);

//			int l=la.size();
			int r=ra.size();
			int u=union.size();

			SH_SW_LA_RA[s0Pos][s1Pos]=new int[u+2];  int a=0;
			SW_LA_RA[s0Pos][s1Pos]   =new int[u+1];  int b=0;
			SH_LA_RA[s0Pos][s1Pos]   =new int[u+1];  int c=0;
			LA_RA[s0Pos][s1Pos]      =new int[u];    int d=0;
			SH_RA[s0Pos][s1Pos]      =new int[r+1];  int e=0;
			RA[s0Pos][s1Pos]         =new int[r];    int f=0;
			
			SH_SW_LA_RA[s0Pos][s1Pos][a++]=sh;
			SH_SW_LA_RA[s0Pos][s1Pos][a++]=sw;
			SW_LA_RA[s0Pos][s1Pos][b++]   =sw;
			SH_LA_RA[s0Pos][s1Pos][c++]   =sh;
			SH_RA[s0Pos][s1Pos][e++]      =sh;
			
			for(Integer i:union){
				SH_SW_LA_RA[s0Pos][s1Pos][a++]=i.intValue();
				SW_LA_RA[s0Pos][s1Pos][b++]=i.intValue();
				SH_LA_RA[s0Pos][s1Pos][c++]=i.intValue();
				LA_RA[s0Pos][s1Pos][d++]=i.intValue();
			}
			for(Integer i:ra){
				SH_RA[s0Pos][s1Pos][e++]=i.intValue();
				RA[s0Pos][s1Pos][f++]=i.intValue();
			}
			
			if(SH_SW_LA_RA[s0Pos][s1Pos].length!=a)
				throw new Error("here1");
			if(SW_LA_RA[s0Pos][s1Pos].length!=b)
				throw new Error("here2");
			if(SH_LA_RA[s0Pos][s1Pos].length!=c)
				throw new Error("here3");
			if(LA_RA[s0Pos][s1Pos].length!=d)
				throw new Error("here4");
			if(SH_RA[s0Pos][s1Pos].length!=e)
				throw new Error("here5");
			if(RA[s0Pos][s1Pos].length!=f)
				throw new Error("here6");
		}
		
	}

	@Override
	public int[] getAllowedTransitions(State s, Instance inst) {
		boolean canArc=s.hasMinTwoStackElements();
		if(!canArc)
			return SH;
		boolean canLA=s.peekStack(1)!=0; //can't LA if root is s1
		boolean canShift=s.bufferSize()>0;
		boolean canSwap=s.tail.s0 < s.s0 && s.tail.s0>0;
		
		int s0=s.peekStack(0);
		int s1=s.peekStack(1);
		int s0Pos=inst.tag[s0];
		int s1Pos=inst.tag[s1];

		
		final int[] r;
		
		if(canLA){
			if(canShift && canSwap)
				r=SH_SW_LA_RA[s0Pos][s1Pos];
			else if(canShift)
				r=SH_LA_RA[s0Pos][s1Pos];
			else if(canSwap)
				r=SW_LA_RA[s0Pos][s1Pos];
			else
				r=LA_RA[s0Pos][s1Pos];
		} else {
			if(canShift)
				r=SH_RA[s0Pos][s1Pos];
			else
				r=RA[s0Pos][s1Pos];
		}
		
		if(r!=null)
			return r;

		//Otherwise use defaults
		
		if(canLA){
			if(canShift && canSwap)
				return d_SH_SW_LA_RA;
			else if(canShift)
				return d_SH_LA_RA;
			else if(canSwap)
				return d_SW_LA_RA;
			else
				return d_LA_RA;
		} else {
			if(canShift)
				return d_SH_RA;
			else
				return d_RA;
		}
	}

}
