package ims.trans.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.AbstractStaticOracle;
import ims.oracle.OracleFactory;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.util.DataSetStats;
import ims.util.IOUtils;

public class SameTRSeqAsOracle {

	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		List<List<String>> oldTRSeq=new ArrayList<List<String>>();
		List<List<String>> newTRSeq=new ArrayList<List<String>>();
		List<List<String>> optTRSeq=new ArrayList<List<String>>();
		
		AbstractStaticOracle oldOracle=OracleFactory.getStaticOracle(TransitionSystem.ArcSOldSwap);
		AbstractStaticOracle newOracle=OracleFactory.getStaticOracle(TransitionSystem.ArcSNewSwap);
		AbstractStaticOracle optOracle=OracleFactory.getStaticOracle(TransitionSystem.ArcSOptimalSwap);
		
		SentenceReader<?> reader=ReaderWriterFactory.getTrainReader(options);
		
		int totalUniqueSeq=0;
		int totalUniqueSeqNoArcAmbig=0;
		
		for(ISentence s:reader){
			oldOracle.initSentence(s.getGoldHeads());
			List<String> oldSeq=oldOracle.getCanonicalSequence(s, true);
			newOracle.initSentence(s.getGoldHeads());
			List<String> newSeq=newOracle.getCanonicalSequence(s, true);
			optOracle.initSentence(s.getGoldHeads());
			List<String> optSeq=optOracle.getCanonicalSequence(s, true);
			
			oldTRSeq.add(oldSeq);
			newTRSeq.add(newSeq);
			optTRSeq.add(optSeq);
			if(DataSetStats.hasUniqueTrSeq(s.getGoldHeads()))
				totalUniqueSeq++;
			if(DataSetStats.hasUniqueTrSeqNoArcAmbig(s.getGoldHeads()))
				totalUniqueSeqNoArcAmbig++;
		}
		
		int[] oldIdentical=new int[options.iterations];
		int[] newIdentical=new int[options.iterations];
		int[] optIdentical=new int[options.iterations];
		int[] allIdentical=new int[options.iterations];
		int[] sameAsLast=new int[options.iterations];
		
		List<List<String>> previous=null;
		
		for(int i=0;i<options.iterations;++i){
			if(!Options.octave)
				System.out.println("Iter: "+i);
			File trFile=TRSeqPathAnalysis.getFileName(options.trSeqFile.toString(), i);
			BufferedReader r=IOUtils.getBufferedReader(trFile, IOUtils.DEFAULT_ENCODING, true);
			List<String> l=IOUtils.slurpLines(r);
			List<List<String>> sorted=sortAndFix(l);
			if(sorted.size()!=oldTRSeq.size())
				throw new Error("length mismatch -- "+sorted.size()+" vs. "+oldTRSeq.size());
			for(int j=0;j<sorted.size();++j){
				boolean o=oldTRSeq.get(j).equals(sorted.get(j));
				boolean n=newTRSeq.get(j).equals(sorted.get(j));
				boolean p=optTRSeq.get(j).equals(sorted.get(j));
				if(o)
					oldIdentical[i]++;
				if(n)
					newIdentical[i]++;
				if(p)
					optIdentical[i]++;
				if(o && n && p)
					allIdentical[i]++;
			}
			if(previous!=null){
				sameAsLast[i]=countSame(previous,sorted);
			}
			previous=sorted;
		}
		
		int totalSen=oldTRSeq.size();

		System.out.printf("TOT_ITER = %d;\n",options.iterations);
		System.out.printf("TOT_SEN = %d;\n",totalSen);

		System.out.println();
		
		System.out.println(Options.octave?String.format("UNIQUE_SEQ = %d;", totalUniqueSeq):String.format("Unique seq: %d",totalUniqueSeq));
		System.out.println(Options.octave?String.format("UNIQUE_SEQ_SW_ONLY = %d;", totalUniqueSeqNoArcAmbig):String.format("Unique seq no arc ambig: %d",totalUniqueSeqNoArcAmbig));
		
		System.out.println();
		
		System.out.print(Options.octave?"SAME_AS_BEFORE = [ ":"Same as last:\n");
		for(int i=0;i<oldIdentical.length;++i){
			double p=100.0*sameAsLast[i]/totalSen;
			System.out.printf("%6.3f ",p);
		}
		System.out.println(Options.octave?"];":"");
		
		System.out.print(Options.octave?"SAME_AS_OLD = [ ":"Old identical:\n");
		for(int i=0;i<oldIdentical.length;++i){
			double p=100.0*oldIdentical[i]/totalSen;
			System.out.printf("%6.3f ",p);
		}
		System.out.println(Options.octave?"];":"");
		
		System.out.print(Options.octave?"SAME_AS_NEW = [ ":"New identical:\n");
		for(int i=0;i<oldIdentical.length;++i){
			double p=100.0*newIdentical[i]/totalSen;
			System.out.printf("%6.3f ",p);
		}
		System.out.println(Options.octave?"];":"");
		
		System.out.print(Options.octave?"SAME_AS_OPT = [ ":"Opt identical:\n");
		for(int i=0;i<oldIdentical.length;++i){
			double p=100.0*optIdentical[i]/totalSen;
			System.out.printf("%6.3f ",p);
		}
		System.out.println(Options.octave?"];":"");
		
		System.out.print(Options.octave?"SAME_AS_ALL = [ ":"All identical:\n");
		for(int i=0;i<allIdentical.length;++i){
			double p=100.0*allIdentical[i]/totalSen;
			System.out.printf("%6.3f ",p);
		}
		System.out.println(Options.octave?"];":"");
	}

	private static int countSame(List<List<String>> previous,List<List<String>> sorted) {
		int r=0;
		for(int i=0;i<previous.size();++i){
			if(previous.get(i).equals(sorted.get(i)))
				++r;
		}
		return r;
	}

	private static List<List<String>> sortAndFix(List<String> l) {
		Comparator<String> cmp=new Comparator<String>(){
			@Override
			public int compare(String o1, String o2) {
				int x1=Integer.parseInt(o1.substring(0, o1.indexOf(':')));
				int x2=Integer.parseInt(o2.substring(0, o2.indexOf(':')));
				return Integer.compare(x1, x2);
			}
		};
		Collections.sort(l, cmp);
		List<List<String>> r=new ArrayList<List<String>>();
		
		for(String s:l){
			List<String> q=Arrays.asList(s.substring(s.indexOf(':')+1).split(","));
			r.add(q);
		}
		return r;
	}
	
}
