package ims.trans.util;

import java.io.IOException;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import gnu.trove.list.TDoubleList;
import gnu.trove.list.array.TDoubleArrayList;
import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.trans.Decoder;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.State;
import ims.trans.transitions.ITransition;
import ims.util.ZipUtil;

public class ScoreTransitions {

	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		Model m=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		Decoder d=new Decoder(m);
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.test, options.format);
		
		for(ISentence sent:reader){
			State state=d.parse(sent, m.beamSize);
			countScores(state);
		}
		
		//Now print avgs
		printScores(basicScores);
		System.out.println();
		System.out.println();
		printScores(fullScores);
	}

	
	private static void printScores(Map<String, TDoubleList> map) {
		for(Entry<String,TDoubleList> e:map.entrySet()){
			String lbl=e.getKey();
			TDoubleList list=e.getValue();
			double sum=list.sum();
			double avg=sum/list.size();
			System.out.printf("%-10s  %8f\n",lbl,avg);
		}
		
	}


	static Map<String,TDoubleList> fullScores=new TreeMap<String,TDoubleList>();
	static Map<String,TDoubleList> basicScores=new TreeMap<String,TDoubleList>();
	
	private static void countScores(State state) {
		while(state.previous!=null){
			double score=state.totalScore-state.previous.totalScore;
			ITransition tr=state.lastAction;
			addScores(score,tr);
			state=state.previous;
		}
	}

	private static void addScores(double score, ITransition tr) {
		addScore(tr.getFullTransitionLabel(),fullScores,score);
		addScore(tr.getBasicTransitionLabel(),basicScores,score);
	}


	private static void addScore(String key, Map<String, TDoubleList> map,double score) {
		TDoubleList list=map.get(key);
		if(list==null){
			list=new TDoubleArrayList();
			map.put(key, list);
		}
		list.add(score);	
	}
	
	
}
