package ims.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class ArrayFunctions {
	
	public static int getHighestIdx(double[] score){
		if(score.length==0)
			throw new Error("empty array???");
		int best=0;
		double bestScore=score[0];
		for(int i=1;i<score.length;++i){
			if(score[i]>bestScore){
				best=i;
				bestScore=score[i];
			}
		}
		return best;
	}
	
	public static int[] integerListToIntArray(List<Integer> l){
		int[] r=new int[l.size()];
		int q=0;
		for(Integer i:l)
			r[q++]=i.intValue();
		return r;
	}
	
	public static int max(int[] a){
		int max=a[0];
		for(int i=1;i<a.length;++i)
			if(a[i]>max)
				max=a[i];
		return max;
	}
	
	public static double averageFromCountArray(int[] arr){
		double sum=0;
		int items=0;
		for(int i=0;i<arr.length;++i){
			items+=arr[i];
			sum+=i*arr[i];
		}
		return sum/items;
	}
	
	public static void shiftAllLeftByBits(long[] arr,int bits,int max){
		for(int i=0;i<max;++i)
			arr[i]<<=bits;
	}
	
	//If we have an array of counts, find the median
	public static double medianFromCountArray(int[] arr){
		int items=0;
		for(int i:arr)
			items+=i;
		
		//This can be done much more efficiently, but I do it the stupid way by constructing the intermediate array
		int[] expanded=new int[items];
		for(int i=0,e=0;i<arr.length;++i){
			for(int p=0;p<arr[i];++p)
				expanded[e++]=i;
		}
		
		if(items%2==1) {//odd number of items, then it's the value "at" items/2
			return expanded[items/2];
		} else {        //else its the mean of the values "at" items/2 and items/2-1
			double sum=expanded[items/2]+expanded[items/2-1];
			return sum/2.d;
		}
	}
	
	public static boolean contains(int[] array,int containee){
		for(int i:array)
			if(i==containee)
				return true;
		return false;
	}

	public static boolean same(int[] a,int[] b){
		if(a==null || b==null)
			return false;
		if(a.length!=b.length)
			return false;
		Arrays.sort(a);
		Arrays.sort(b);
		for(int i=0;i<a.length;++i)
			if(a[i]!=b[i])
				return false;
		return true;
	}
	
	public static int countSame(int[] i,int[] j){
		return countSame(i,j,0,i.length);
	}
	
	public static int countSame(int[] i, int[] j,int from,int until) {
		if(i==null || j==null)
			return 0;
		if(i.length!=j.length)
			throw new Error("errr");
		int same=0;
		for(int q=from;q<until;++q){
			if(i[q]==j[q])
				++same;
		}
		return same;
	}
	
	public static int countSame(float[] i,float[] j){
		return countSame(i,j,0,i.length);
	}
	
	public static int countSame(float[] i, float[] j,int from,int until) {
		if(i==null || j==null)
			return 0;
		if(i.length!=j.length)
			throw new Error("errr");
		int same=0;
		for(int q=from;q<until;++q){
			if(i[q]==j[q])
				++same;
		}
		return same;
	}
	
	public static <T> boolean sameTwice(T[] n){
		Set<T> seen=new HashSet<T>();
		for(T a:n){
			if(seen.contains(a))
				return true;
			seen.add(a);
		}
		return false;
	}
	
	public static <T> int countNonNull(T[] a){
		int cnt=0;
		for(T t:a)
			if(t!=null)
				++cnt;
		return cnt;
	}
	
	@SafeVarargs
	public static <T> List<T> concat(List<T>... lists){
		List<T> r=new ArrayList<T>();
		for(List<T> l:lists)
			r.addAll(l);
		return r;
	}
}
