package ims.util.visualtrans;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.imageio.ImageIO;

import com.googlecode.whatswrong.Edge;
import com.googlecode.whatswrong.NLPCanvasRenderer;
import com.googlecode.whatswrong.NLPInstance;
import com.googlecode.whatswrong.SingleSentenceRenderer;
import com.googlecode.whatswrong.Token;
import com.googlecode.whatswrong.TokenProperty;

public class WhatsWrongExport {

	public static final String DEPTYPE_BLACK="dep";
	public static final String DEPTYPE_RED="depRED";
	public static final String DEPTYPE_BLUE="depBLUE";
	public static final String DEPTYPE_PURPLE="depPURPLE";
	
	
	private static final TokenProperty PROP_FORM=new TokenProperty("Form", 0);
	private static final TokenProperty PROP_ID  =new TokenProperty("ID", 1);
	
	NLPCanvasRenderer renderer=new SingleSentenceRenderer();
	
	public WhatsWrongExport(){
		renderer.setEdgeTypeColor(DEPTYPE_BLACK, Color.BLACK);
		renderer.setEdgeTypeColor(DEPTYPE_RED, Color.RED);
		renderer.setEdgeTypeColor(DEPTYPE_BLUE, Color.BLUE);
		renderer.setEdgeTypeColor(DEPTYPE_PURPLE, Color.MAGENTA);
	}
	
    private static BufferedImage getImage(NLPCanvasRenderer renderer,NLPInstance instance,double scaleFactor){
    	BufferedImage image=new BufferedImage(1, 1, BufferedImage.TYPE_4BYTE_ABGR);
    	Graphics2D graphics=image.createGraphics();
    	Dimension d=renderer.render(instance, graphics);
    	
    	if(scaleFactor!=1)
    		d=new Dimension((int) (d.width*scaleFactor),(int) (d.height*scaleFactor));
    	
    	image=new BufferedImage((int) d.getWidth(),(int) d.getHeight(),BufferedImage.TYPE_4BYTE_ABGR);
    	graphics=image.createGraphics();
    	if(scaleFactor!=1)
    		graphics.scale(scaleFactor, scaleFactor);
    	
    	renderer.render(instance, graphics);
    	return image;
    }
    
    public void exportToJPG(File file,NLPInstance instance,double scaleFactor) throws FileNotFoundException, IOException{
    	BufferedOutputStream os=new BufferedOutputStream(new FileOutputStream(file));
    	exportToJPG(os,instance,scaleFactor);
    	os.close();
    }
    
    /**
     * Export to JPG. For some reason the background comes out pink, I have no idea why.
     * Note that this method /does not/ close the outputstream after writing, this is up to the caller.
     * 
     * @param os The stream to write to
     * @param renderer The renderer to use
     * @param instance The instance to render
     * @param scaleFactor The scaling factor. Set to 1 for default size.
     * @throws IOException 
     */
    public void exportToJPG(OutputStream os,NLPInstance instance,double scaleFactor) throws IOException{
    	BufferedImage image=getImage(renderer,instance,scaleFactor);
    	ImageIO.write(image,"JPG",os);
    }
    
    public void exportToPNG(File file,NLPInstance instance,double scaleFactor) throws FileNotFoundException, IOException{
    	BufferedOutputStream os=new BufferedOutputStream(new FileOutputStream(file));
    	exportToPNG(os,instance,scaleFactor);
    	os.close();
    }
    
   /**
    * Export to PNG.
    * Note that this method /does not/ close the outputstream after writing, this is up to the caller.
    *  
    * @param os The stream to write to
    * @param renderer The renderer to use
    * @param instance The instance to render
    * @param scaleFactor The scaling factor. Set to 1 for default size.
    * @throws IOException 
    */
    public void exportToPNG(OutputStream os,NLPInstance instance,double scaleFactor) throws IOException{
    	BufferedImage image=getImage(renderer,instance,scaleFactor);
    	ImageIO.write(image,"PNG",os);	
    }
	
	public static NLPInstance getNLPInstance(String[] forms,int[] heads,boolean withID,String[] depTypes){
		List<Token> tokens=new ArrayList<Token>();
		List<Edge> edges=new ArrayList<Edge>();
		for(int i=0,size=heads.length;i<size;++i){
			tokens.add(new Token(i));
		}
		
		for(int i=0,size=heads.length;i<size;++i){
			Token t=tokens.get(i);
			String form=forms[i];
			if(form==null || form.equals(""))
				form="!";
			t.addProperty(PROP_FORM,form);
			if(withID)
				t.addProperty(PROP_ID,Integer.toString(i));
			int tokenHead=heads[i];
			if(tokenHead>=0){

				String type="dep";
				if(depTypes!=null)
					type=depTypes[i];
//				if(depsToFlag!=null && depsToFlag.contains(i))
//					type="depCOLOR";
				edges.add(new Edge(tokens.get(tokenHead),t," ",type));
			}
		}
		NLPInstance instance=new NLPInstance(tokens,edges,NLPInstance.RenderType.single,Collections.<Integer> emptyList());
		return instance;
	}
    
	
	public void setCurved(boolean curved){
		renderer.setCurved(curved);
	}
	
	public void setFlagDepsColor(Color c){
		renderer.setEdgeTypeColor("depCOLOR", c);
	}
}
